package Atlas::Scaffold;
$VERSION=2.0;

use strict;
use Atlas::Utility::ObjectAttribute;
use Atlas::Project::Trace;
use Atlas::Project::Contig;
use IO::File;
use Heap;
use Atlas::ScaffoldHeapEle;

use vars qw($MAX_INSERT_SIZE $VERBOSE $MIN_CONTIG_LENGTH $ERROR_RATE $MAX_WEIGTH $ABORT $THRESHOLD);  #global variable
use vars qw(%path_table @edges $WGSDENSITY $MIN_GOOD_PLACEMENT $PAIR_MODE $DB_NAME $DB_TRACE_MODE $DB_PAIR_MODE $HGSC_DB);
use vars qw($DB_TRACE_TABLE1 $DB_TRACE_TABLE2 $DB_PAIR_TABLE $BUNDLE_SIZE);

$MIN_GOOD_PLACEMENT=100;   #the minimal good read placement in a contig to be trustable
$MAX_INSERT_SIZE=1000000;  # 1 Mb, should be large enough for any kind of insertion, except YAC
$VERBOSE=0;
$MIN_CONTIG_LENGTH=800;
$ERROR_RATE=0.2;
$MAX_WEIGTH=10000000;
$ABORT=0;
$THRESHOLD=5; #the minmal number of pair clones found before overwrite the external data
$WGSDENSITY=6; #make the wgs density at 6 so that have half to a third weight as a skim pair
$PAIR_MODE=1; #default will care pair not read persa, pairs that fall within a contig will only reflect as template
$DB_NAME=''; #the default database
$DB_TRACE_MODE=0; #default not to load table, 1,2 for different table and 3 for table2 but write to file first
$DB_TRACE_TABLE1='unpair_read';
$DB_TRACE_TABLE2='all_asm_read';
$DB_PAIR_TABLE='pair_template'; 
$DB_PAIR_MODE=0; #defualt not to load table, print to a file instead
$BUNDLE_SIZE=3000;  #default windown size to bundle edges

=pod
=head1 Atlas::Scaffold
    Scaffold.pm is a similar package of BacProject.pm .  However, this package focus on global scaffold and
    will implement different algorithm in getting the scaffold.  At the end, this package will replace the
    BacProject.pm .

    The features that will be included in this package are:
    1. ability to choose send all the info to a database or keep them in memory
    2. more checking the pair reads correctness in terms of their placement in contig.  Mainly rely on triming
    3. better judgement of the edges between two contigs.  calculate the good count, negative count and fail count
    4. better estimation of the gap size.  Libraries with different insert size will be consided sequencially to get
       a better estimation of the gap size and reduce the error rate.

=cut

#------------------------------------
#global variables accessor and mutator
#------------------------------------

sub set_error_rate {
    my $self=shift;
    my($new_value)=shift;

    $ERROR_RATE=$new_value if(defined $new_value);
    return $ERROR_RATE;
}

sub get_path_table_ref {
    return \%path_table;
}

sub get_edges_array_ref {
    return \@edges;
}

sub set_max_insert_size {
    my($new_value)=shift;

    $MAX_INSERT_SIZE=$new_value if(defined $new_value);
    return $MAX_INSERT_SIZE;
}

sub set_min_contig_length {
    my $self=shift;
    my($new_value)=shift;

    $MIN_CONTIG_LENGTH=$new_value if(defined $new_value);
    return $MIN_CONTIG_LENGTH;
}


sub set_verbose {
    my $self=shift;
    my($new_value)=shift;

    $VERBOSE=$new_value if(defined $new_value);
    return $VERBOSE;
}

sub set_abort {
    my($new_value)=shift;

    $ABORT=$new_value if(defined $new_value);
    return $ABORT;
}

sub set_pair_mode {
    my($new_value)=shift;

    $PAIR_MODE=$new_value if(defined $new_value);
    return $PAIR_MODE;
}

sub set_db_name {
    my($new_value)=shift;

    $DB_NAME=$new_value if(defined $new_value);
    return $DB_NAME;
}

sub set_db_trace_mode {
    my($new_value)=shift;

    $DB_TRACE_MODE=$new_value if(defined $new_value);
    return $DB_TRACE_MODE;
}

sub set_db_pair_mode {
    my($new_value)=shift;

    $DB_PAIR_MODE=$new_value if(defined $new_value);
    return $DB_PAIR_MODE;
}

sub set_db_tracetable1 {
    my($new_value)=shift;

    $DB_TRACE_TABLE1=$new_value if(defined $new_value);
    return $DB_TRACE_TABLE1;
}

sub set_db_tracetable2 {
    my($new_value)=shift;

    $DB_TRACE_TABLE2=$new_value if(defined $new_value);
    return $DB_TRACE_TABLE2;
} 
 
sub set_db_pairtable {
    my($new_value)=shift;

    $DB_PAIR_TABLE=$new_value if(defined $new_value);
    return $DB_PAIR_TABLE;
}

sub set_bundle_size {
    my $self=shift;
    my($new_value)=shift;
    $BUNDLE_SIZE=$new_value if(defined $new_value);
    
    return $BUNDLE_SIZE;
}

sub get_score_array {
    return (50, 875, 1700, 2025, 2350, 2475, 2600, 2650, 2675, 2700, 2725);
}

#----------------------------------------------------
#accessor and mutator function generated on the fly
#----------------------------------------------------

my @_project_attribute=(
			'report_type',     #file type
			'report_file',     #file name
			'mincontiglength', #minimal lenght of contig to be processed
			'wholegenome', #the mode of the project
			'db_handle', #database handle object
			'trace_table_type', #trace table type
			'trace_sth',       #trace table loading statement handle
			'pair_sth',
			'name', #id/name of the project
 			);
&Atlas::Utility::ObjectAttribute::declare_rw(@_project_attribute);

###
#constructor
##########
sub create {
    my($class)=shift;

    my($project)=+{
	'contigs'=>[],
	'samples'=>{},
	'graph'=>{},
	'delete_sample'=>[],
    };
    bless($project,$class);
}

=head1 prepare_sth


=cut
sub prepare_trace_sth {

    my $self=shift;
    my $dbh=shift;

    $dbh=$self->db_handle() if(!defined $dbh);

    my $cmd;
    #based on the table type prepare sth
    if($DB_TRACE_MODE == 1) { #it is the unpair_read table
	$cmd="insert into $DB_TRACE_TABLE1
                     (trace_name, contig_name, trace_start, trace_end, strand, direction, sample, library)
                    values(?,?,?,?,?,?,?,?)";
    } elsif($DB_TRACE_MODE == 2) {
	$cmd="insert into $DB_TRACE_TABLE2
                     (trace_name, project_name, contig_name)
                     values(?,?,?)";
    }
    my $sth;
    $sth=$dbh->prepare($cmd) if(defined $cmd);

    #save the sth
    $self->trace_sth($sth);
}

=head1 prepare_pair_sth


=cut
sub prepare_pair_sth {

    my $self=shift;
    my $dbh=shift;

    my $cmd;
    #based on the table type prepare sth
    if($DB_PAIR_MODE == 1) { #it is the unpair_read table
	$cmd="insert into  $DB_PAIR_TABLE
                     (contig_name, sample, library, startpos, endpos, clonesize)
                    values(?,?,?,?,?,?)";
    }
    my $sth;
    $sth=$dbh->prepare($cmd) if(defined $cmd);

    #save the sth
    $self->pair_sth($sth);
}

#clean up the project sample hash, which is used to find all the intercontig pairs
sub Reset_Project_Sample {
    my $self=shift;
    
    delete $self->{'samples'};
}

#clean up the samples that link two contigs
sub Reset_Inter_Contig_Sample {
    my $self=shift;

    delete $self->{'project_pair_samples'};
}

#clean up the samples that link two contigs
sub Unify_Delete_Sample {
    my $self=shift;

    my %seen=();
    my @uniq=();

    foreach my $sample (@{$self->{'delete_sample'}}) {
	my $template=$sample->[0]->sample();
	push @uniq, $sample unless (exists $seen{$template});
	$seen{$template}++;
    }
    
    @{$self->{'delete_sample'}}=();
    push @{$self->{'delete_sample'}}, @uniq;
    
}

#there are circular reference between contigs and traces back to project
sub DESTORY {

    my $self=shift;
    
    #destory all back link from contigs to project
    my @contigs=$self->list_contigs();
    foreach my $contig (@contigs) {
	delete $contig->{'project'};
	#do the same to all traces
	my @traces=$contig->list_traces();
	foreach my $trace (@traces) {
	    delete $trace->{'contig'};
	    delete $trace->{'project'};
	}
    }
    
    $self->db_handle()->disconnect() if($self->db_handle());

}
#--------------------------------------------------------------------
# Create a contig from the project
# So to establish the aggregation relationship between project and contig
#---------------------------------------------------------------------- 
sub _create_contig {
    my($self)=shift;
    my $contig=Atlas::Project::Contig->create();
    $contig;
}


#---------------------------------
# add a new contig into the project
#----------------------------------
sub add_contig {
    my($self)=shift;
    my($contig)=shift;

    # store the project package name into the contig object
    $contig->project($self);
    push(@{$self->{'contigs'}},$contig);
}


#----------------------------------------------------------
# Return a list of the contigs in the current project object
#----------------------------------------------------------
sub list_contigs {
    my($self)=shift;

    @{$self->{'contigs'}};
}

#-----------------------------------------------------------
#return the total length of all contigs in the object
#if an argument is passed, it is used as the minimal length
#only contigs that longer than this number will be calculated
#without an argument, all contigs will be calculated
#------------------------------------------------------------
sub all_contig_length {
    my($project)=shift;
    my($min_length)=shift;
    my($length, $contig, $contig_length);

    $min_length=0 unless (defined $min_length);
    $length=0;
    foreach $contig (@{$project->{'contigs'}}) {
	$contig_length=$contig->length();
	if($contig_length > $min_length) {
	    $length += $contig_length;
	}
    }

    return $length;
}


#---------------------------------------------------------
#create a new trace object
#---------------------------------------------------------
sub _create_trace {
    my($self)=shift;
    my($tracename)=shift;

    my $trace=Atlas::Project::Trace->create($tracename);
    $trace;
}

#---------------------------------------------------------
#Purpose: add a trace object  ref to the project, not directly, but in 
# a hash using the template name as the key.  At the mean time,
# mark the trace $trace->project with the caller project
#Precondition: Call from a project, pass the trace ref
#---------------------------------------------------------
sub add_trace {
    my($self)=shift;
    my($trace)=shift;

    $trace->project($self);
    push(@{$self->{'samples'}{$trace->mapgap_group()}},$trace);
}

#-------------------------------------------------------------
#Purpose: return a list of all traces ref from a give template
#Precodintion: call from a project, pass the template name
#Post: nothing return if no such template exist
#------------------------------------------------------------
sub list_traces_for_sample {
    my($self)=shift;
    my($sample)=shift;

    @{$self->{'samples'}{$sample}};
}

#---------------------------------------------------------
#Purpose: return the sample hash
#Pre: call from a project with has finishing add all samples
#Return value: a hash reference that with the sample name as key 
# and an array of trace objects as the value. Not duplication,
# the same copy as the project object, be careful
#-------------------------------------------------------
sub list_all_samples {
    my($self)=shift;
    
    return $self->{'samples'};
}

=head1 read_file
    
   In: an assembly file
   Return: Nothing
    
   Note: Adaptors/Readers are provided for ace.old, ace.new, xgap and taf(tab-deliminated assembly file) format
    Normally, for a local scaffold, a local assembly ace file is provided.  For a global assembly, a taf file is
    generated from a database and supplied to the scaffold object to read.

=cut

sub read_file {
    my($self)=shift;
    my($filename)=shift;
    my($outfh)=shift;

    my($firstline);

    #open the file base on the file type
    if($filename=~/\.gz$/) { #gz file
	open(FIN, "gunzip -c $filename |") || die "Couldnot open file $filename\n";
    } else {
	# Read the first line of the file.
	open(FIN, "$filename") || die "Couldnot open file $filename\n";
    }

    while ($firstline=<FIN>) {
	last unless $firstline=~/^\s*$/; #break when not empty line
    }
    #rewind
    #seek FIN, 0, 0;  how to?
    close FIN;

    if($filename=~/\.gz$/) { #gz file
	open(FIN, "gunzip -c $filename |") || die "Couldnot open file $filename\n";
    } else {
	# Read the first line of the file.
	open(FIN, "$filename") || die "Couldnot open file $filename\n";
    }

    if(defined $outfh) { #a linear stuff
	$self->_read_newace_file_linear(\*FIN, $outfh);
    }
    # Guess what kind of file it is.
    elsif ($firstline=~/^DNA /) {
	$self->_read_ace_file($filename, \*FIN);
    }
    elsif ($firstline=~/^( Total|===)/) {
	print STDERR "Read of xgap relation file\n";
	$self->_read_xgap_file($filename, \*FIN);
    }
    elsif ($firstline=~/^AS/) {
	print STDERR "Read of the new ace file\n";
	$self->_read_newace_file($filename, \*FIN);
    }
    elsif( $firstline=~/^TAF/) {
	print STDERR "Read of a TAF file for global assembly\n";
	$self->_read_taf_file($filename, \*FIN);
    }
    else {
	print "$filename has a unknown format, please check\n";
    }

    close FIN;
}

=head1 _read_ace_file
  Purpose: Parse an old format ace report
  IN: the ace file handle for reading  
  POST: all the information have been saved into the object
   contigs are built into contig objects and stored in 
   the array project->{'contig'}
=cut
sub _read_ace_file {
    my($self)=shift;
    my $filename=shift;
    my($fh)=shift;

    my($line, $curr_contig, $trace);

    # Record some info.
    $self->report_file($filename);
    $self->report_type('ace');

    while ($line=<$fh>) {

	# Start of a contig
	# DNA Contig1
	if ($line=~/^DNA\s+(Contig\d+)/ || $line=~/DNA\s+(proj_\S+)/) {
	    $curr_contig=$self->_create_contig();
	    $curr_contig->name($1);
	    $self->add_contig($curr_contig);
	    #
	    next;
	}

	# Trace data
	# Assembled_from* TA1_1711.s  1  749
	if ($line=~/^Assembled_from\*\s+(.+?)\s+([-\d]+)\s+([-\d]+)/o) {
	    $trace=$self->_create_trace($1);
	    $trace->contig_start($2);
	    $trace->contig_end($3);
            $trace->contig($curr_contig);
	    $curr_contig->add_trace($trace);  #add this trace to the contig
	    $self->add_trace($trace);   #add this trace to the project
	    next;
	}
    }
}

=head1 _read_newace_file

  Purpose: Parse the ace file with new format
  IN: New format ace file handle
  OUT: all the infor stored in the object

  Note: DS line stuff can be handled and over write the parsing based on read name

=cut
sub _read_newace_file {
    my($self)=shift;
    my $filename = shift;
    my($fh)=shift;

    my($line, $curr_contig, $trace);
    my %tracehash=(); #a hash to keep tracking traces name in the project, key is the trace name
                      #full string and value is the trace object reference

    # Record some info.
    $self->report_file($filename);
    $self->report_type('ace2');
     
    my $curr_read_name;
    my $pre_read_name;
    while ($line=<$fh>) {
	# Start of a contig
	# CO Contig1 341 1 1 U
	if ($line=~/^CO\s+(Contig\S+)\s+(\d+)/ || $line=~/^CO\s+(proj_\S+)\s+(\d+)/) { #for unique naming of contig, fix
                                                                                     # with format proj_projname_contig1
	    $curr_contig=$self->_create_contig();
	    $curr_contig->name($1);
	    $self->add_contig($curr_contig);
	    $curr_contig->length($2);
	    $curr_contig->end($2);
	    $curr_contig->start(1);

	    #delete the last read from the last contig from the trace hash
	   if(defined $curr_read_name && exists $tracehash{$curr_read_name}) {
	       if(defined $pre_read_name && exists $tracehash{$pre_read_name}) {
		   my $pretraceref=$tracehash{$pre_read_name};
		   #get the previous read position
		   my $pre_start=$pretraceref->contig_start();
		   my $pre_end=$pretraceref->contig_end();
		   
		   my $curtraceref=$tracehash{$curr_read_name};
		   my $cur_start=$curtraceref->contig_start();
		   my $cur_end=$curtraceref->contig_end();

		  # print "$pre_start\t$pre_end\t$cur_start\t$cur_end\n";
		   unless ($pretraceref->contig() eq $pretraceref->contig()) {
		       delete $tracehash{$curr_read_name};
		   }
		   if($pre_end-$cur_start<350 || $cur_end-$pre_end>200) {
		     #  print "delete\n";
		       delete $tracehash{$curr_read_name};
		   }
	       } else {
		   delete $tracehash{$curr_read_name};
	       }
	    }

	    next;
	}

	#read in the contig quality data to trim the contig size
	if($line=~/^BQ/) {
	    my($cl, $cr);
	    $cl=0;
	    $cr=0;
	    my @base_qual=();
	    #advance the line until it reaches to an empty line
	    while($line=<$fh>) {
		chomp($line);
		#chop off leading and trailing space if any
		$line =~ s/^(\s+)//;
		$line =~ s/(\s+)$//;

		last if $line=~/^\s*$/;
		my @data=split /\s+/, $line;
		push @base_qual, @data;
	    }

	    #count from the start and the end of the array to get the trim
	    for(my $i=0; $i<=$#base_qual; $i++) {
		last if($base_qual[$i] != 0);
		$cl++;
	    }
	    for(my $i=$#base_qual; $i>=0; $i--) {
		last if($base_qual[$i] != 0);
		$cr++;
	    }
	    $curr_contig->start($cl);
	    my $cr_end=$curr_contig->length()-$cr;
	    $cr_end=0 if($cr_end<0);
	    $curr_contig->end($cr_end);
	    if($cr_end-$cl+1<=1) {
		$curr_contig->length(0);
	    } else {
		$curr_contig->length($cr_end-$cl+1);
	    }
	}
	    
		  
       # Trace data
       # AF XBM23456.x1.scf U 1
	if ($line=~/^AF\s+(.+?)\s+([a-z])\s+(-{0,1}\d+)/io) {
	    next if ($3<0);
	    $trace=$self->_create_trace($1);
	    $trace->contig_start($3);
	    $trace->contig_end(0);
            $trace->contig($curr_contig);
	    if($2 eq 'C') {
		$trace->contig_strand('-');
	    } else {
		$trace->contig_strand('+');
	    }
	    $tracehash{$1}=$trace;   #add the trace reference to the hash inorder to update its contig_end later
	    next;
	}

	#get the contig_end of a trace
	if($line=~/RD\s+(.+?)\s+(\d+)/o) {
	    if( exists $tracehash{$1} ) {
		my $end=$tracehash{$1}->contig_start()+$2;
		$tracehash{$1}->contig_end($end);

		$pre_read_name=$curr_read_name;

		$curr_read_name=$1;
	    } else { #this read is rejected
		undef $curr_read_name;
	    }
	    next;
	}
	
	#adjust the trace read placement based on quality
	#QA   1       450        3           445
	#  qual_left  qual_r    align_l      align_r 
	if($line=~/^QA\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/) {
	    if( defined $curr_read_name && exists $tracehash{$curr_read_name} ) {
		my $curtraceref=$tracehash{$curr_read_name};
		#delete this read if no good placement better than 100bp found
		my ($left, $right);
		if($1<$3) {
		    $left=$3;
		} else {
		    $left=$1;
		}

		if($2<$4) {
		    $right=$2;
		} else {
		    $right=$4;
		}
	#	print STDERR "delete $curr_read_name for low quality placement\n" if ($right-$left<$MIN_GOOD_PLACEMENT);
		delete $tracehash{$curr_read_name} if($right-$left<$MIN_GOOD_PLACEMENT);
		
		#adjust the read position in the contig to the good alignment region
		my $start=$curtraceref->contig_start()+$3-1;
		my $end=$curtraceref->contig_start()+$4-1;
		$curtraceref->contig_start($start);
		$curtraceref->contig_end($end);
	    }
	    next;
	}

	#get more info from the trace from the DS line
	if($line=~/^DS/) {
	    if( defined $curr_read_name && exists $tracehash{$curr_read_name} ) { 
                #get the file name, which should be uniqe
		my $curtraceref=$tracehash{$curr_read_name};
		$curtraceref->sample($1) if ( $line =~ /TEMPLATE:\s+(\S+)/i );
		if( $line =~ /DIRECTION:\s+(\S+)/i ) {
		    if ($1 eq 'fwd') {
			$curtraceref->direction('F');
		    } else {
			$curtraceref->direction('R');
		    }
		} elsif($line=~/CENTER: BCM/) { #a baylor read, hard code to fix error upstream
		    $line=~/^DS\s*(\S+)\s*/;
		    my $realtrace=$self->_create_trace($1);
		    $curtraceref->direction($realtrace->direction());
		}

		if ( $line =~ /LIBRARY:\s+(\S+)/i ) {
		    $curtraceref->coslib($1);
		    unless (exists $self->{'contig_pair_samples'}->{$1}) {
			my $libname=$1;
			my $libsize=$1 if($line =~ /INSERT_SIZE:\s(\d+)/i);
			my $libstd=$1 if($line =~ /INSERT_STDEV:\s(\d+)/i);
			my $libinfo={'exp_average_size'=>$libsize,
				     'exp_stddev'=>$libstd};
			$self->add_new_library($libname, $libinfo);
		    }			
		} elsif($line =~ /INSERT_SIZE:\s(\d+)/ && $line =~ /CENTER:\s(\S+)/) { #try to accomodate missing libname
		    my $centername=$1 if($line =~ /CENTER:\s(\S+)/i);                    #use centername . $libsize as lib
		    my $libsize=$1 if($line =~ /INSERT_SIZE:\s(\d+)/i);
		    my $libstd=$1 if($line =~ /INSERT_STDEV:\s(\d+)/i);
		    my $libname=$centername . $libsize;
		    my $libinfo={'exp_average_size'=>$libsize,
				 'exp_stddev'=>$libstd};
		    $self->add_new_library($libname, $libinfo);
		}
	    }
	    next;
	}
		  
    } #end of while

    #begin to add all the traces in the %tracehash to the contigs and project
    while( my($key, $value) = each(%tracehash) ) {
	$curr_contig=$value->contig();
	next if (! $value->mapgap_group()); #skip if the template or library is not defined
	$curr_contig->add_trace($value);  #add this trace to the contig
	$self->add_trace($value);   #add this trace to the project
    }
}

=head1 _read_newace_file_linear

  Purpose: Parse the information from new ace file one contig a time to save space and provide
           the possibility of parallel
  IN: ace file handle
  OUT: print out taf file while read through the ace file one contig a time

  Note: The information it print out is essentially the placement of all reads assembled in the contig
       All these data can be stored into a database table for future access.

=cut
sub _read_newace_file_linear {
    my($self)=shift;
    my ($fh)=shift;    #ace file handle
    my ($outfh)=shift; #out put file handle

    my($line, $curr_contig);
    my %tracehash=(); #a hash to keep tracking traces name in the project, key is the trace name
                      #full string and value is the trace object reference

    my %lib_info=(); #store all the predetermined lib info

    my $curr_read_name;
    my $pre_read_name;
    while ($line=<$fh>) {
	# Start of a contig
	# CO Contig1 341 1 1 U
	if ($line=~/^CO\s+(Contig\S+)\s+(\d+)/ || $line=~/^CO\s+(proj_\S+)\s+(\d+)/) { #for unique naming of contig, fix
                                                                                     # with format proj_projname_contig
		
	    #now one contig is finished, print out all the traces that are not paired in this contig
	    #calculate the paired reads and the error ones(-1 for size)
	    if(defined $curr_contig) {

		#skip the contig if the size of the contig is less than min_length
		if($curr_contig->length() < $MIN_CONTIG_LENGTH) {
		    undef %tracehash;
		} else {
		    my $contig_name=$curr_contig->name();
		    #sort the hash based on contig_start
		    my @trace_names=sort{$tracehash{$a}->contig_start() <=> $tracehash{$b}->contig_start() } keys %tracehash;

		    #delete the last read from the last contig from the trace hash
		    if(defined $curr_read_name && exists $tracehash{$curr_read_name}) {
			if(defined $pre_read_name && exists $tracehash{$pre_read_name}) {
			    my $pretraceref=$tracehash{$pre_read_name};
			    #get the previous read position
			    my $pre_start=$pretraceref->contig_start();
			    my $pre_end=$pretraceref->contig_end();
		   
			    my $curtraceref=$tracehash{$curr_read_name};
			    my $cur_start=$curtraceref->contig_start();
			    my $cur_end=$curtraceref->contig_end();

			    # print "$pre_start\t$pre_end\t$cur_start\t$cur_end\n";
			    if($pre_end-$cur_start<350 || $cur_end-$pre_end>200) {
				#  print "delete\n";
				delete $tracehash{$curr_read_name};
			    }
			} else {
			    delete $tracehash{$curr_read_name};
			}
		    }


		    $self->find_pair_reads(\%tracehash, \@trace_names, $contig_name, $outfh) if($PAIR_MODE); 
		                                                                 #want to get pair,do not care read persa

		    #now print out those unpaired ones
		    foreach my $trace_name (@trace_names) { 
			next unless exists $tracehash{$trace_name}; 
			my $trace_ref=$tracehash{$trace_name};	
			next if(!$trace_ref->mapgap_group());
			my $trace_start=$trace_ref->contig_start();
			my $trace_end=$trace_ref->contig_end();
			my $trace_strand=$trace_ref->contig_strand();
			my $trace_dir=$trace_ref->direction();
			my $trace_sample=$trace_ref->sample();
			my $trace_lib=$trace_ref->coslib();
			if(!defined $trace_dir || !defined $trace_sample || !defined $trace_lib) {
			    print "WARNING: $trace_name unparsed\n";
			} else {
			    print $outfh "TRACE\t$trace_name\t$contig_name\t$trace_start\t$trace_end\t$trace_strand\t$trace_dir\t$trace_sample\t$trace_lib\n" if(!$DB_TRACE_MODE);
			    #load the trace info into the right table
			    $self->load_trace_table($trace_name, $contig_name, $trace_start, $trace_end, 
						    $trace_sample, $trace_strand, $trace_dir, $trace_lib, $outfh) if($DB_TRACE_MODE);
			}
			#release the trace object
			delete $tracehash{$trace_name};
		    }
		}
	    }

	    #start with a new contig
	    $curr_contig=$self->_create_contig();
	    $curr_contig->name($1);	
	    $curr_contig->length($2);
	    $curr_contig->start(1);
	    $curr_contig->end($2);
	    next;
	}

	#read in the contig quality data to trim the contig size
	if($line=~/^BQ/) {
	    my($cl, $cr);
	    $cr=$cl=0;
	    my @base_qual=();
	    #advance the line until it reaches to an empty line
	    while($line=<$fh>) {
		chomp($line);

		#chop off leading and trailing space if any
		$line =~ s/^(\s+)//;
		$line =~ s/(\s+)$//;

		last if $line=~/^\s*$/;
		my @data=split /\s+/, $line;
		push @base_qual, @data;
		
	    }
	    #count from the start and the end of the array to get the trim
	    for(my $i=0; $i<=$#base_qual; $i++) {
		last if($base_qual[$i] != 0);
		$cl++;
	    }
	    for(my $i=$#base_qual; $i>=0; $i--) {
		last if($base_qual[$i] != 0);
		$cr++;
	    }
	    $curr_contig->start($cl);
	    my $cr_end=$curr_contig->length()-$cr;
	    $cr_end=0 if ($cr_end<0);
	    $curr_contig->end($cr_end);

	    #adjust length, 0 as minimal
	    if($cr_end-$cl+1>=0) {
		$curr_contig->length($cr_end-$cl+1);
	    } else {
		$curr_contig->length(0);
	    }

	    #print out the contig information
	    my $contig_len=$curr_contig->length();
	    my $contig_name=$curr_contig->name();
	    my $contig_start=$curr_contig->start();
	    my $contig_end=$curr_contig->end();
	    print $outfh "CONTIG\t$contig_name\t$contig_start\t$contig_end\t$contig_len\n";
	}
	    
		  
       # Trace data
       # AF XBM23456.x1.scf U 1
	if ($line=~/^AF\s+(.+?)\s+([a-z])\s+(-{0,1}\d+)/io) {
	    next if ($3<0);
	    my $trace=$self->_create_trace($1);
	    $trace->contig_start($3);
	    $trace->contig_end(0);
	    if($2 eq 'C') {
		$trace->contig_strand('-');
	    } else {
		$trace->contig_strand('+');
	    }
	    $tracehash{$1}=$trace;   #add the trace reference to the hash inorder to update its contig_end later
	    next;
	}

	#get the contig_end of a trace
	if($line=~/RD\s+(.+?)\s+(\d+)/o) {
	    if( exists $tracehash{$1} ) {
		my $end=$tracehash{$1}->contig_start()+$2;
		$tracehash{$1}->contig_end($end);

		$pre_read_name=$curr_read_name;

		$curr_read_name=$1;
	    } else { #this read is rejected
		undef $curr_read_name;
	    }
	    next;
	}
	
	#adjust the trace read placement based on quality
	#QA   1       450        3           445
	#  qual_left  qual_r    align_l      align_r 
	if($line=~/^QA\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/) {
	    if( defined $curr_read_name && exists $tracehash{$curr_read_name} ) {
		my $curtraceref=$tracehash{$curr_read_name};
		#delete this read if no good placement better than 100bp found
		my ($left, $right);
		if($1<$3) {
		    $left=$3;
		} else {
		    $left=$1;
		}

		if($2<$4) {
		    $right=$2;
		} else {
		    $right=$4;
		}
		
	#	print "delete $curr_read_name for low quality placement\n" if ($right-$left<$MIN_GOOD_PLACEMENT);
		delete $tracehash{$curr_read_name} if($right-$left<$MIN_GOOD_PLACEMENT);
		
		#adjust the read position in the contig to the good alignment region
		my $start=$curtraceref->contig_start()+$3-1;
		my $end=$curtraceref->contig_start()+$4-1;
		$curtraceref->contig_start($start);
		$curtraceref->contig_end($end);
	    }
	    next;
	}

	#get more info from the trace from the DS line
	if($line=~/^DS/) {
	    if( defined $curr_read_name && exists $tracehash{$curr_read_name} ) { 
                #get the file name, which should be uniqe
		my $curtraceref=$tracehash{$curr_read_name};
		$curtraceref->sample($1) if ( $line =~ /TEMPLATE:\s+(\S+)/i );
		if( $line =~ /DIRECTION:\s+(\S+)/i ) {
		    if ($1 eq 'fwd') {
			$curtraceref->direction('F');
		    } else {
			$curtraceref->direction('R');
		    }
		} elsif($line=~/CENTER: BCM/) { #a baylor read, hard code to fix error upstream
		    $line=~/^DS\s*(\S+)\s*/;
		    my $realtrace=$self->_create_trace($1);
		    $curtraceref->direction($realtrace->direction());
		}

		if ( $line =~ /LIBRARY:\s+(\S+)/i ) {
		    $curtraceref->coslib($1);
		    unless (exists $lib_info{$1}) {
			my $libname=$1;
			my $libsize=$1 if($line =~ /INSERT_SIZE:\s(\d+)/i);
			my $libstd=$1 if($line =~ /INSERT_STDEV:\s(\d+)/i);
			if(defined $libsize && defined $libstd) {
			    print $outfh "LIB\t$libname\t$libsize\t$libstd\n";
			    $lib_info{$libname}++;
			}
		    }
		}
	    }
	    next;
	}
		  
    } #end of while

    #print out the rest of stuff in the tracehash if the contig is good size
    return if(!defined $curr_contig || $curr_contig->length() < $MIN_CONTIG_LENGTH);

    #delete the last read from the last contig from the trace hash
    if(defined $curr_read_name && exists $tracehash{$curr_read_name}) {
	if(defined $pre_read_name && exists $tracehash{$pre_read_name}) {
	    my $pretraceref=$tracehash{$pre_read_name};
	    #get the previous read position
	    my $pre_start=$pretraceref->contig_start();
	    my $pre_end=$pretraceref->contig_end();
	    
	    my $curtraceref=$tracehash{$curr_read_name};
	    my $cur_start=$curtraceref->contig_start();
	    my $cur_end=$curtraceref->contig_end();
	    
	    # print "$pre_start\t$pre_end\t$cur_start\t$cur_end\n";
	    if($pre_end-$cur_start<350 || $cur_end-$pre_end>200) {
		#  print "delete\n";
		delete $tracehash{$curr_read_name};
	    }
	} else {
	    delete $tracehash{$curr_read_name};
	}
    }
    
    my $contig_name=$curr_contig->name(); #the last one
    my @trace_names=sort{$tracehash{$a}->contig_start() <=> $tracehash{$b}->contig_start() } keys %tracehash;
    $self->find_pair_reads(\%tracehash, \@trace_names, $contig_name, $outfh) if($PAIR_MODE); #want to get pair,do not care read pers
    foreach my $trace_name (@trace_names) {
	next unless exists $tracehash{$trace_name}; 
	my $trace_ref=$tracehash{$trace_name};
	next if (! $trace_ref->mapgap_group()); 
	my $trace_start=$trace_ref->contig_start();
	my $trace_end=$trace_ref->contig_end();
	my $trace_strand=$trace_ref->contig_strand();
	my $trace_dir=$trace_ref->direction();
	my $trace_sample=$trace_ref->sample();
	my $trace_lib=$trace_ref->coslib();
	print $outfh "TRACE\t$trace_name\t$contig_name\t$trace_start\t$trace_end\t$trace_strand\t$trace_dir\t$trace_sample\t$trace_lib\n" if(! $DB_TRACE_MODE);
	$self->load_trace_table($trace_name, $contig_name, $trace_start, $trace_end, 
				$trace_sample, $trace_strand, $trace_dir, $trace_lib, $outfh) if($DB_TRACE_MODE);
	#release the trace object
	delete $tracehash{$trace_name};
    }

}

=head1 load_pair_table


=cut
sub load_pair_table {
    my $self=shift;

    my($contig_name, $sample, $library, $startpos, $endpos, $clonesize)=@_;
    
    my $sth=$self->pair_sth();
    
    $sth->bind_param(1, $contig_name);
    $sth->bind_param(2, $sample);
    $sth->bind_param(3, $library);
    $sth->bind_param(4, $startpos);
    $sth->bind_param(5, $endpos);
    $sth->bind_param(6, $clonesize);

    print "$contig_name, $sample, $library, $startpos, $endpos, $clonesize\n" unless($sth->execute());
}
=head1 load_trace_table


=cut
sub load_trace_table {
    my $self=shift;

    my($trace_name, $contig_name, $trace_start, $trace_end, $trace_sample, $trace_strand, $trace_dir, $trace_lib, $outfh)=@_;

    #based on purpose, get the right sth
    my $sth=$self->trace_sth();

    #prepare the data, if detail one(pair/trace)
    if($DB_TRACE_MODE == 1) {
	$sth->bind_param(1, $trace_name);
	$sth->bind_param(2, $contig_name);
	$sth->bind_param(3, $trace_start);
	$sth->bind_param(4, $trace_end);
	$sth->bind_param(5, $trace_strand);
	$sth->bind_param(6, $trace_dir);
	$sth->bind_param(7, $trace_sample);
	$sth->bind_param(8, $trace_lib);
    } elsif($DB_TRACE_MODE ==2) {
	my $project_name=$self->name();
	#only load wgs reads 
	return if($trace_name =~ /^$project_name/i);
	$sth->bind_param(1, $trace_name);
	$sth->bind_param(2, $project_name);
	$sth->bind_param(3, $contig_name);
    } elsif($DB_TRACE_MODE==3) {
	my $project_name=$self->name();
	#only load wgs reads 
	return if($trace_name =~ /^$project_name/i);
	print $outfh "$trace_name\t$project_name\t$contig_name\t$trace_start\t$trace_end\t$trace_sample\t$trace_strand\t$trace_dir\t$trace_lib\n";
	#print "$trace_name\n" if (!defined $trace_name || !defined $project_name || !defined $contig_name || !defined $trace_start);
	#print "$trace_name\n" if (!defined $trace_dir);
	return;
    }

    $sth->execute();
}

=head1 find_pair_reads

  Purpose: Find all the read pair in the same contig and test whether they are correct
  In: a set of trace objects that belong to the same contig defined by contig_name
  OUT: pair reads found will be elimimated from the trace object set

=cut
sub find_pair_reads {

    my $self=shift;
    my($tracehash_ref, $tracename_ref, $contig_name, $outfh)=@_;

    my %pair=();
    foreach my $trace_name (@{$tracename_ref}) {
	my $trace_ref=$tracehash_ref->{$trace_name};
	next unless defined $trace_ref;
	next if (! $trace_ref->mapgap_group());
	if(exists $pair{$trace_ref->mapgap_group()}) { #the pair have been seen
	    #add another one
	    push @{$pair{$trace_ref->mapgap_group()}}, $trace_ref;
	    #delete them from the trace hash
	    delete $tracehash_ref->{$trace_name};
	    my $read1=$pair{$trace_ref->mapgap_group()}->[0];
	    delete $tracehash_ref->{$read1->original_str()};
	}
	push @{$pair{$trace_ref->mapgap_group()}}, $trace_ref;
    }

    #now for a pair, check and save
    foreach my $group (keys %pair) {
	    #see if it is a good pair
	next if($#{$pair{$group}}<1); #skip if there is only one read in it
	my $read1=$pair{$group}->[0];
	my $strand1=$read1->contig_strand();
	my $trace_sample=$read1->sample();
	my $trace_lib=$read1->coslib();

	my $read2=$pair{$group}->[-1];
	my $strand2=$read2->contig_strand();
	my $start=$read1->contig_start();
	my $end=$read2->contig_end();
	if( ($strand1 eq '-') || ($strand1 eq $strand2) ) {
	    print $outfh "PAIR\t$contig_name\t$trace_sample\t$trace_lib\t$start\t$end\t-1\n" if(!$DB_PAIR_MODE);
	    $self->load_pair_table($contig_name, $trace_sample, $trace_lib, $start, $end, -1) if($DB_PAIR_MODE);
	} else { #a good pair
	    my $size=$end-$start;
	    print $outfh "PAIR\t$contig_name\t$trace_sample\t$trace_lib\t$start\t$end\t$size\n" if(!$DB_PAIR_MODE);
	    $self->load_pair_table($contig_name, $trace_sample, $trace_lib, $start, $end, $size) if($DB_PAIR_MODE);
	}
    }

}


# Read an xgap show-relationships report.
#
#  CONTIG            NUMBER  LENGTH               ENDS
#                                              LEFT   RIGHT
#                        45      820            190     247
#  NAME              NUMBER POSITION LENGTH     NEIGHBOURS
#                                              LEFT   RIGHT
#  I61U0518.xdap        190        1   -239       0     665
#  I61U1325.xdap        665        2   -419     190     396

sub _read_xgap_file {
    my($self)=shift;
    my $filename = shift;
    my($fh)=shift;


    my($contig_count, $curr_contig, $line, $trace); #

    $contig_count=1;

    # Record some info.
    $self->report_file($filename);
    $self->report_type('xgap');

    #open(FIN,$filename);

    #
    while ($line=<$fh>) {
	chomp($line);

	# Xgap 3:==========
	# Total contig length     55140.  Average length     1225.3
	# Total characters in gel readings    258034.
	# Average gel characters per consensus character       4.68
	# Number of gel readings    695 Number of contigs     45
	#
	# CONTIG            NUMBER  LENGTH               ENDS
	#                                             LEFT   RIGHT
	#                        0      820            190     247
	# NAME              NUMBER POSITION LENGTH     NEIGHBOURS
	#                                             LEFT   RIGHT
	#
	# Xgap4:==========
	# ============================================================
	# Mon 28 Jul 15:16:23 1997: show relationships
	# ------------------------------------------------------------
	
	# Skip over headings like above.
	if ($line=~/^\s+(CONTIG|NAME|LEFT|Total|Number|Average|GEL)/ or
	    $line=~/^\s*$/ or
	    $line=~/^[-=]+$/ or
	    $line=~/show relationships/ or
	    $line=~/close database/ or
	    $line=~/join contigs/ or
	    $line=~/edit contig/ or
	    $line=~/complement contig/ or
	    $line=~/Complementing contig/
	    ) {
#	    print "header: $line\n";
	    next;
	}

	# The start of a contig has 4 numbers
	if ($line=~/^\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$/) {
#	    print "Contig: $1\n";
	    $curr_contig=$self->_create_contig();
	    $curr_contig->name($contig_count++);
	    $curr_contig->length($2);
	    $self->add_contig($curr_contig);
	    next;
	}

	# name number pos len left right
	if ($line=~/^\s*([^\s]+)\s+(\d+)\s+(\d+)\s+(-?\d+)\s+(\d+)/) {
	    if (!defined($curr_contig)) {
		die "no contig!!!\n";
	    }
	    $trace=$self->_create_trace($1);
	    if ($trace->format()) {
		$trace->{'xgap_id'}=$2;
		$trace->contig_start($3);
		$trace->contig_end($3+abs($4));
                $trace->contig($curr_contig);
		if ($4>0) {
		    $trace->contig_strand('+');
		}
		else {
		    $trace->contig_strand('-');
		}
		$curr_contig->add_trace($trace);
		$self->add_trace($trace);
	    }
	    else {
		print "     | Skipping trace '$line'\n";
	    }
	    next;
	}
	print "Error: Unable to understand input!\n";
	print "Look at ",$line," of the input file.\n";
	print ">>>",$_,"\n";
	exit;
    }
}

=head1 cal_contig_pair

  Purpose: calculate the pair-end sequenced clone properties,
    including the density, insertion size (maxi, min and average).
  Precondition: Call from a project, pass the minimal contig length for skipping contigs
  Other: the error rate for deciding if throw a library is 20%, it is hard coded in the
    global variable.  The minimal contig length is 2000bp, also hard coded in the global
    variable.
  Error: the following read pair will be throw out from the analysis:
    1. the most left and right sequence are same direction in the contig
    2. clone size less than 0
    3. libraries that more than 20% of the sequences have been throw out.  For this case,
    the error rate of this library will stay at 1.  So, should check the error rate
    before work on a library statistics.
  Notes:
    The way stddev is calculated is directly on the read length.  If there are a lot of 
    read pairs, the sum of square(length) can be huge.  Can take the precision at 100bp to 
    reduce the sum or use big float data type.  Not implemented yet
=cut


sub cal_contig_pair {

    my($project)=shift;
    my($min_contig_len)=shift; #the minimal size of the contig will be calculated

    my($contig, $sample, $trace, @contig_pair_samples, $library, $library_name);
    my($left, $right, $clone_size, $total_length, $average, $stddev, $sample_count, $contig_length, @ordered);
    my($max, $min, $error_rate, $error_count, $sq_sum, $sum, $pass_count);

    $min_contig_len=$MIN_CONTIG_LENGTH unless (defined $min_contig_len);  #the default minimal length is 800bp

    foreach $contig (@{$project->{'contigs'}}) {

        if($contig->length() < $min_contig_len) {
	    print "Contig: $contig->{'name'} is smaller than $min_contig_len, skip.\n" if ($VERBOSE);
            next;
        }
      
        #********************************************************************************
        # traverse each large contig and save pair end sequences into the project like:
	#  $Project = +{
	#                'contig_pair_samples'=>{ 'coslib1' => { 'count' => NUMBER,
	#                                                       'max_size' => NUMBER,
	#                                                       'min_size' => NUMBER,
	#                              this step  ------->      'samples' => [] (paired sample traces array)
	#                                                     },
	#                                          'coslib2' => ...
	#                                        }
        #              }
	#**********************************************************************************
	
        @contig_pair_samples=$contig->list_pair_samples();

        foreach $sample (@contig_pair_samples) {
	    $library_name=$sample->[0]->coslib();
	    push(@{$project->{'contig_pair_samples'}->{$library_name}->{'samples'}}, $sample);
	}

    }

    #
    #calculate the statistics for each clone library based on pair_end samples from
    #that library
    #
    my $contig_total_length=$project->all_contig_length($min_contig_len);

    while( ($library_name, $library)= each(%{$project->{'contig_pair_samples'}}) ){ #use which here in case huge hash
	$error_count=0;
	$sample_count=$#{$library->{'samples'}}+1;
        $average=$stddev=0;
        $error_rate=1;
        $total_length=0;
	$min=$MAX_INSERT_SIZE;
	$max=0;
	$sq_sum=$sum=0;

	foreach $sample (@{$library->{'samples'}}) {  # traverse the trace array for each template            
             $left=$right=0;
             @ordered = sort { $a->contig_start() <=> $b->contig_start() } @$sample;

             if($ordered[0]->{'contig_strand'} eq '+' && $ordered[-1]->{'contig_strand'} eq '-') {
		 $left=$ordered[0]->{'contig_start'};
                 $right=$ordered[-1]->{'contig_end'};
                 $clone_size=$right-$left;

                 if($clone_size > 0) {
		     $total_length += $clone_size;
                     if($clone_size > $max) { $max=$clone_size;}
                     if($clone_size < $min) { $min=$clone_size;}
		     $sq_sum += $clone_size * $clone_size;
                 }
                 else {  #thorw out case 2
                     $error_count++;
                     printf STDOUT ("%s Discarded for negative size\n", $sample->[0]->mapgap_group() ) if ($VERBOSE);
                 }
              }

	      else {  #throw out case 1
                  $error_count++;
                   printf STDOUT ("%s Discarded for not pair\n", $sample->[0]->mapgap_group() ) if ($VERBOSE);
	      }
         }
        
	$pass_count=$sample_count-$error_count;
        $error_rate=$error_count/$sample_count unless $sample_count==0;
                                                  # calculate the percentage of discarded templates
        if($error_rate > $ERROR_RATE ) { 
	    print "something is wrong with library: $library_name, error rate greater than $ERROR_RATE\n" if($VERBOSE && $pass_count>$THRESHOLD);
	    $library->{'error_rate'}=$error_rate;
	    $library->{'average_size'}=0;
	    $library->{'max_size'}=0;
	    $library->{'min_size'}=0;
	    $library->{'density'}=0;  # Number of pairs per kb region
	    $library->{'stddev'}=0;
	    $library->{'total_pair'}=$sample_count;
	    $library->{'pass_pair'}=$pass_count;
	    next;
	}

        $average=$total_length/$pass_count unless $pass_count<=0;
	if($pass_count>1) { #standard deviation
	    $stddev=sqrt( ($sq_sum - $pass_count*$average*$average)/($pass_count-1) );
	}

        $library->{'average_size'}=$average;
        $library->{'max_size'}=$max;
        $library->{'min_size'}=$min;
        $library->{'error_rate'}=$error_rate;
        $library->{'density'}=$pass_count*1000/$contig_total_length;  # Number of pairs per kb region
	$library->{'stddev'}=$stddev;
        $library->{'total_pair'}=$sample_count;
	$library->{'pass_pair'}=$pass_count;

    } #end of while loop

    $project->_reconcile_libraries(); #an internal function that reconcile external libraries and calculated results

} #end of sub

#---------------------------
#change the library properties
# $self->set_library_prop(libname, \%argu)
# argument hash should be {'average_size'=>$average;
  #                         'max_size'}=$max;
  #                         'min_size'}=$min;
  #                         'error_rate'}=$error_rate;
  #                          'density'
#	                     'stddev'}=$stddev;
#                            'total_pair'}=$sample_count;
#	                     'pass_pair'}=$pass_count;
#---------------------------
sub set_library_prop {

    my $project=shift;

    my $lib_name=shift;
    my $arg=shift;

    #get the library ref
    #for each property in the hash, change it
    my $lib;
    if(exists $project->{'contig_pair_samples'}{$lib_name}) {
        $lib=$project->{'contig_pair_samples'}{$lib_name};
    } else {
	$project->{'contig_pair_samples'}{$lib_name}={};
	$lib=$project->{'contig_pair_samples'}{$lib_name};
    }

    #populate the arguments 
    foreach my $key (keys %{$arg}) {
	$lib->{$key}=$arg->{$key};
    }
}

#---------------------------------
#reconcile_libraries
#---------------------------------
sub _reconcile_libraries {
    my $project=shift;
    
    my ($library_name, $library, %lib_hash);
    while( ($library_name, $library)= each(%{$project->{'contig_pair_samples'}}) ){ #use which here in case huge hash
	if($library->{'pass_pair'}<=$THRESHOLD) { #only a few reads from this library has been found, so stick with external data
	    $library->{'average_size'}=$library->{'exp_average_size'} if(defined $library->{'exp_average_size'});
	    $library->{'stddev'}=$library->{'exp_stddev'} if(defined $library->{'exp_stddev'});
	    if(exists $library->{'exp_average_size'}) { #wgs reads, so it is normal no pair found internally
		$library->{'density'}=$WGSDENSITY;
		$library->{'error_rate'}=0;
	    }
	}
    }
}
#-----------------------------------------------------------------------------------------
#Purpose: access function that return a hash that contains info for all libraries
#   of the project
#-------------------------------------------------------------------------------------------
sub get_library_info {
    my $project=shift;
    my ($library_name, $library, %lib_hash);

    %lib_hash=();
    while( ($library_name, $library)= each(%{$project->{'contig_pair_samples'}}) ){ #use which here in case huge hash
	while( my($key, $value)= each (%{$library}) ) {
	    next if ($key eq 'samples');
	    $lib_hash{$library_name}->{$key}=$value;
	}
    }
    return \%lib_hash;
}

#--------------------------------------------------------
#add_new_library
#add external information of a new library
#input: caller is the project, the first is the library name
#       the second is the hash reference of the library, currently
#       only expect 'exp_average_size' and 'exp_stddev'
#POST: the library has been added to the 'contig_pair_samples'
#------------------------------------------------------
sub add_new_library {
    my $project=shift;
    my $library_name=shift;
    my $infohashref=shift;

    while( my($key, $value)=each %{$infohashref} ) {
	$project->{'contig_pair_samples'}->{$library_name}->{$key}=$value;
    }
}

#--------------------------------------------------------------------
#Purpose: Find all pair end samples that assembled in more than
#  one contig and save the template name into array $project->{'project_pair_samples'}
#Pre: called from a project.  Also make sure all samples have been
# current and added to the project
#Post: all pair samples added to the @{$project->{'project_pair_samples'}}
#Note: In the array, not necessary pair ends.  If some templates have been
# walked, for instant, and assembled into different contigs, these also
# will be found and saved.  However, only ends that fall into two contigs will
# be selected.  If a templated has been sequenced all the way and assembled into
# more than two contigs, those will not been considered pair ends.  Different
# sub can be developed to deal with these situations
#Error: No rejection will be made this step
#----------------------------------------------------------
sub find_inter_pair {
    my($project)=shift;
    my ($min_size, $max_size)=@_;

    $min_size=0 unless (defined $min_size);
    $max_size=10000000 unless (defined $max_size);

    my($sample, $traces, $contig, $contig_name, $found);
    my($i, %seen);

    $found=0;
    #get the lib info
    my($library_name, $library);
    my($start_node, $end_node, $max_overlap, $density, $weigth, $stddev);
    
    #identify bad libraries
    my %bad_lib=();
    while( ($library_name, $library)= each(%{$project->{'contig_pair_samples'}}) ){ #use which here in case huge hash
	if($library->{'error_rate'}>$ERROR_RATE || $library->{'average_size'}<$min_size 
	   || $library->{'average_size'}>$max_size ) {
	    $bad_lib{$library_name}++;
	}
    }

    #group all paired reads based on the two nodes they connect
OUTER:    while( ($sample, $traces)= each(%{$project->{'samples'}}) ) {

        if($#$traces < 1) { next; } #$trace is really an array of traces from the same sample

	#if the trace come from bad lib or not lib info, skip
	$library_name=$traces->[0]->coslib();
	next if exists $bad_lib{$library_name};
	next unless exists $project->{'contig_pair_samples'}->{$library_name};

        %seen=();
        for($i=0; $i <= $#$traces; $i++) {
	    $contig=$traces->[$i]->contig();
	    next OUTER if ($contig->length() < $MIN_CONTIG_LENGTH);
            $contig_name=$contig->name();
            $seen{$contig_name}++;
	}

	if( keys (%seen) == 2) {  #only traces assemblled into two contigs will be considered
             #push(@{$project->{'project_pair_samples'}}, $traces);  #save space, speed
	     #put them into group based on the nodes they link
	     my ($contig1, $contig2)=keys %seen;
	     ($contig1, $contig2)=sort ($contig1, $contig2);
	     push (@{$project->{'project_pair_samples'}->{$contig1}->{$contig2}}, $traces);
	     $found++;
	     #print "$sample\n";
        }
    } #end of while loop

    print "total of $found new pairs found\n";

} #end of sub

#-------------------------------------------------------------------------------
# Purpose: Find all trustable inter-contig pair-end sequence and save the graph
# Pre: The statistics of each library have been calculated based on intra-contig
# Post: all good edges have been stored with relative scores
# Algorithm:
#     step 1: bundle the egdgs between two contigs and calculate the gap size and count
#     step 2: add all edge boundles(>=2) to the graph
#
#    'graph'=>{'node_name1'=>{
#                              'next_node_name'=>{
#                                                 'gap'=>
#                                                 'samples'=>{ 'library_name'=>{'count'=>
#                                                                               'left_arm'=>
#                                                                               'right_arm'=>
#                                                                               'samples'=>[]
#                                                                              }
#                                                            }
#                                                 }
#                             }
#               'node_name2'=>{
#
#                             } 
#             }
#
#         for each pair sample {
#              for both of the contig the pair sequence is in
#                 next if the read orientation is not the same in the same contig
#                 find the clone end in each contig(the farest point from the contig end
#                                                   based on orientation)                             
#                 next if minimal clone size > cut off value
#                 store it in the graph:
#                  if it contain new node
#                      create new node
#                      create new edge
#                  if it contain old node
#                     update the edge
#                     record the lefest arm and rightes arm based on which library this sample from
#         }
#         for each edge {
#             calculate the confidence
#         }
#          
#                 
#       
#-------------------------------------------------------------------------------------------

sub build_graph {
    my($project)=shift;
    my $pre_exists=shift;

    my($i, $sample, $trace, $flag, $contig, $contig_len, $contig_name, $count, $clone_end);

    #init the scaffold and contig lookup table
    $project->init_scaffold() unless (defined $pre_exists);

    #now for all samples that connect two contigs, seperate them into scaffolds and caculate the edge
    my ($node1, $node2); #two connected node by the edge
    my %error_pair=();
    my %bundle_edge=();
    foreach my $contig1_name (keys %{$project->{'project_pair_samples'}}) {
	foreach my $contig2_name (keys %{$project->{'project_pair_samples'}->{$contig1_name}}) {
	    my @samples=@{$project->{'project_pair_samples'}->{$contig1_name}{$contig2_name}};
	    #loop through every sample to the graph
	    #print "add sample to graph\n";
	    $project->add_sample_to_graph(\@samples);
	}
    }

    $project->Cal_Bundle();

} #end of sub

#-----------------------------------------
#$scaffold->init_scaffold()
#To initiate the scaffold link list
# This list is to record all the current linklist and its contig components(start, end)
#-----------------------------------------------------------
sub init_scaffold {

    my $self=shift;

    #get all the contigs from the object, the initial linklist are those contigs only, each one
    # list has only one contig to start with
    my @contigs=$self->list_contigs();
    foreach my $contig (@contigs) {
	next if ($contig->length() < $MIN_CONTIG_LENGTH);
	my $sca_id=$contig->name();
	my $start=1;
	my $end=$contig->length();
	my $orientation='+';
	$self->{'scaffold'}->{$sca_id}->{$sca_id}={'start'=>$start,
						   'end'=>$end,
						   'orientation'=>$orientation};
	
	$self->{'scaffold'}->{$sca_id}->{'size'}=$end;
	$self->{'scaffold'}->{$sca_id}->{'head'}=$sca_id;
	$self->{'scaffold'}->{$sca_id}->{'tail'}=$sca_id;
	$self->{'scaffold'}->{$sca_id}->{'sample'}=[];
	$self->{'contig_scaffold'}->{$sca_id}=$sca_id;  #lookup between contig to scaffold id
    }
}

#------------------------------------------
#$scaffold->Cal_Bundle()
#Input: an array of templates that connect two nodes
#Output: the estimated gap size and template counts that support this
#Algorithm:
#        if there are only one template, calculate the gap size and return
#        foreach template,
#               based on the library info and read positions in each contig, calculate the gap size
#        sort all these template based on the estimated gap size
#        using the bundle width(class varialbe) as the window to find the maxi bundle and get the gap
#        size and the average size, return the count in the window.
#----------------------------------------------------------------
sub add_sample_to_graph {
    my $self=shift;
    my $all_sample=shift; #a ref of array, which contains all templates connecting the two nodes
    my $force=shift;  #if to force add repetitive stuff
    $force=0 unless defined $force;

    my $edges=$self->{'graph'};  #the hash to store all the passed template and the calculated data
    my $rep_node=$self->{'rep_node'};

    my($library_name, $library, $node1, $node2);

OUTER:    foreach my $sample (@{$all_sample}) {
	my %seen=();  # a lexcial variable that always go out of scope each loop for memory reuse
	
        #calculate the size that overalp with both contigs
	my ($clone_end);
	for(my $i=0; $i <= $#{$sample}; $i++) {
	    my $contig=$sample->[$i]->contig();
	    my $contig_name=$contig->name();
	    my $strand=$sample->[$i]->contig_strand();
	    $seen{$contig_name}={ 'strand'=>$strand,
	       		     'contig'=>$contig};
	    
	    #begin to record clone end
	    my $contig_node;
	    if ($strand eq '+') {
		$clone_end = $contig->end() - $sample->[$i]->contig_start();
		$contig_node=$contig_name . '_right';
	    } else {
		$clone_end = $sample->[$i]->contig_end()-$contig->start();
		$contig_node=$contig_name . '_left';
	    }
	 
	    #print "read location: strand $strand " . $sample->[$i]->contig_start() . " contig size is ". $contig->end() ."\n" if ($VERBOSE);
	    #print "$i: $clone_end \n " if ($VERBOSE);
	    #print $sample->[$i]->coslib() . " in $contig_node\n"
	   # if($force==0 && exists $rep_node->{$contig_node}) {
	#	print $sample->[$i]->sample() . " is rejected for landing in repetitive contigs\n" if ($VERBOSE);
	#	next OUTER;
	#    }

	    if( exists $seen{$contig_name}->{'clone_end'} && $seen{$contig_name}->{'clone_end'} > $clone_end) {
		#do nothing
	    } else {
		$seen{$contig_name}->{'clone_end'} = $clone_end;
	    }
	} #end of for loop
	  
	my $key;
	my $clone_arm=0;
	foreach $key (keys %seen) {
	    $clone_arm += $seen{$key}->{'clone_end'};
	}
	
	$library_name=$sample->[0]->coslib();
	  
	#if the library has not been seen intra contig, skip
	if(exists $self->{'contig_pair_samples'}{$library_name}) {
	    $library=$self->{'contig_pair_samples'}{$library_name};
	} else {
	    print " the library: $library_name is not present\n" if ($VERBOSE);
	    next;
	}

	#now tranlate the contig into the scaffold id/pos
	#although the pairs connect the same two contigs, there are orientation difference
	my $template=$sample->[0]->mapgap_group();
	my $contig1=$sample->[0]->contig()->name();
	my $sca1=$self->{'contig_scaffold'}->{$contig1};
	my $sca1_size=$self->{'scaffold'}{$sca1}{'size'};
	
	my $contig2=$sample->[-1]->contig()->name();
	my $sca2=$self->{'contig_scaffold'}->{$contig2};
	my $sca2_size=$self->{'scaffold'}{$sca2}{'size'};

	next if($sca1 eq $sca2);

	#print "scaffold is $sca1 and contig is $contig1\n";
	if($self->{'scaffold'}{$sca1}{$contig1}{'orientation'} eq '+') {
	    if($sample->[0]->contig_strand() eq '+') {
		$node1=$sca1 . '_right';
		$clone_arm += $sca1_size-$self->{'scaffold'}{$sca1}{$contig1}{'end'};
	    } else {
		$node1=$sca1 . '_left';
		$clone_arm += $self->{'scaffold'}{$sca1}{$contig1}{'start'}-1;
	    }
	} else {
	    if($sample->[0]->contig_strand() eq '+') {
		$node1=$sca1 . '_left';
		$clone_arm += $self->{'scaffold'}{$sca1}{$contig1}{'start'}-1;
	    } else {
		$node1=$sca1 . '_right';
		$clone_arm += $sca1_size-$self->{'scaffold'}{$sca1}{$contig1}{'end'};
	    }
	}

	if($self->{'scaffold'}{$sca2}{$contig2}{'orientation'} eq '+') {
	    if($sample->[-1]->contig_strand() eq '+') {
		$node2=$sca2 . '_right';
		$clone_arm += $sca2_size-$self->{'scaffold'}{$sca2}{$contig2}{'end'};
	    } else {
		$node2=$sca2 . '_left';
		$clone_arm += $self->{'scaffold'}{$sca2}{$contig2}{'start'}-1;
	    }
	} else {
	    if($sample->[-1]->contig_strand() eq '+') {
		$node2=$sca2 . '_left';
		$clone_arm += $self->{'scaffold'}{$sca2}{$contig2}{'start'}-1;
	    } else {
		$node2=$sca2 . '_right';
		$clone_arm += $sca2_size-$self->{'scaffold'}{$sca2}{$contig2}{'end'};
	    }
	}

	#print "clone_arm is $clone_arm\n" if($VERBOSE);
	#make a new edge and add to the graph
	my $gap_size=$library->{'average_size'}-$clone_arm;
	
	my $edge={'gap_size'=>$gap_size,
		  'sample'=>$sample
		  };
	push @{$edges->{$node1}->{$node2}{'sample'}}, $edge;
	push @{$edges->{$node2}->{$node1}{'sample'}}, $edge;
    }
}


#-------------------
#print out all deleted samples
#---------------------------
sub print_deleted_sample {
    my $self=shift;

    print "inside deleted smaple\n";
    my $delete=$self->{'delete_sample'};
    foreach my $sample (@{$delete}) {
	print "deleted: " . $sample->[0]->sample() . "\n";
    }
}


#------------------------------
#sub to drop edges/clonepairs that too internal of the scaffold to be true
#use with discreetion
#
#traverse the graph, examine each active link, if the gap size from this link is 
#4*stddev smaller than 0, delete this link
#-----------------------------------------------
sub Clean_Graph {

    my $self=shift;
    
    my $edges=$self->{'graph'};

    foreach my $node1 (keys %{$edges}) {
	foreach my $node2 (keys %{$edges->{$node1}}) {
	    #loop through all the samples
	    next unless defined $edges->{$node1}->{$node2}{'sample'};
	    my @samples=@{$edges->{$node1}->{$node2}{'sample'}};
	    @{$edges->{$node1}->{$node2}{'sample'}}=();
	    foreach my $sample (@samples) {
		#get the lib info for this sample
		my $lib=$sample->{'sample'}->[0]->coslib();
		my $template=$sample->{'sample'}->[0]->sample();
		my $std=$self->{'contig_pair_samples'}{$lib}{'stddev'};
		if($sample->{'gap_size'}>(-4*$std)) {
		    #keep
		    push @{$edges->{$node1}->{$node2}{'sample'}}, $sample;
		} else {
		   # push @{$self->{'delete_sample'}}, $sample->{'sample'}; #already in delete array
		    print "delete $template between $node1 and $node2\n" if ($VERBOSE);
		}
	    }
	    if($#{$edges->{$node1}->{$node2}{'sample'}}<0) {
		delete $edges->{$node1}{$node2};
	    }
	}
    }

}
#----------------------
#sub to caculate the bundlize links and give the
#gap size, max, stddev for each edge based on the samples
#in the edge hash of the object
#without given an specific edge ref, will caculate all active edges of the object
# the edge object has the following structure
#  $ref={'sample'=>[{'gap_size'=>number, 'sample'=>[trace, trace]}.{}]}
#
#--------------------------------------

sub Cal_Bundle {
    my $self=shift;
    my $in_node1=shift;
    my $in_node2=shift;
    
    my $edges;
    if(!defined $in_node1 || !defined $in_node2) {
	$edges=$self->{'graph'};
    } else {
	if($in_node1=~/[_left|_right]$/ && $in_node2=~/[_left|_right]$/) {
	    my $temp=$self->{'graph'}->{$in_node1}->{$in_node2};
	    $edges->{$in_node1}={$in_node2=>$temp};
	} else {
	    $edges->{$in_node1.'_left'}{$in_node2.'_left'}=$self->{'graph'}{$in_node1.'_left'}{$in_node2.'_left'}
	                if(exists $self->{'graph'}{$in_node1.'_left'}{$in_node2.'_left'});
	    $edges->{$in_node1.'_left'}{$in_node2.'_right'}=$self->{'graph'}{$in_node1.'_left'}{$in_node2.'_right'}
	                if(exists $self->{'graph'}{$in_node1.'_left'}{$in_node2.'_right'});
	    $edges->{$in_node1.'_right'}{$in_node2.'_left'}=$self->{'graph'}{$in_node1.'_right'}{$in_node2.'_left'}
	                if(exists $self->{'graph'}{$in_node1.'_right'}{$in_node2.'_left'});
	    $edges->{$in_node1.'_right'}{$in_node2.'_right'}=$self->{'graph'}{$in_node1.'_right'}{$in_node2.'_right'}
	                if(exists $self->{'graph'}{$in_node1.'_right'}{$in_node2.'_right'});
	}
    }

    my $window=&set_bundle_size();
   # print "Window size is $window\n";
    my @score_array=$self->get_score_array();

    my %seen=();
    my $key;
    #now sort the %edges based on the estimated gap size
    foreach  my $node1 (keys %{$edges}) {
	foreach my $node2 (keys %{$edges->{$node1}}) {
	    $key=$node1.'_'.$node2;
	    next if exists $seen{$key};
	    next unless exists $edges->{$node1}->{$node2}{'sample'};
	    my @unsort=@{$edges->{$node1}->{$node2}{'sample'}};
	    next unless $#unsort>=0;

	    #sorting the edge
	    my @sorted=();
	    @sorted=sort {$a->{'gap_size'} <=> $b->{'gap_size'}} @unsort;
	    
	    #calculate bundle, this algorithm is slow,can be speed up in the future
	    my $count=1;
	    my $max=1;
	    my $floor=$sorted[0]->{'gap_size'};
	    my $top=$sorted[-1]->{'gap_size'};
	#    print "$floor\n";
	    my $total_gap=$floor;
	    my $average_gap=$floor;
	    my $floor_index=1;
	    for(my $i=1; $i<=$#sorted; $i++) {
		$top=$sorted[$i]->{'gap_size'};
	#	print "$top\n";
		if($top-$floor<=$window) {
		    $count++;
		    $total_gap += $top;
		    if($i==$#sorted) {#reach the last one
			$average_gap=$total_gap/$count;
			$max=$count;
		    }		  
		} else { #reach the window maxi
		    if($count>$max) { #a better window
			$average_gap=$total_gap/$count;
			$max=$count;
		    }
		    #reset
		    last if $i>=$#sorted-1;
		    $floor_index++;
		    $i=$floor_index;
		    $count=1;
		    $floor=$sorted[$i]->{'gap_size'};
		    $total_gap=$floor;
		}
	    }
	 #   print "ave size $average_gap\n";
	    $self->{'graph'}->{$node1}->{$node2}{'ave_gap_size'}=$average_gap;
	    $self->{'graph'}->{$node1}->{$node2}{'count'}=$max;
	    $self->{'graph'}->{$node1}->{$node2}{'total_count'}=$#sorted+1;

	    my $pass_count=0;
	    my $sq_sum=0;
	    my $stddev;
	    #calculate the weight and the stddev of the connection
	    foreach my $template (@sorted) {
		if( abs($template->{'gap_size'}-$average_gap)<$window ) {
		    $sq_sum += ($template->{'gap_size'}-$average_gap) * ($template->{'gap_size'}-$average_gap);
		    $template->{'pass'}=1;
		    $pass_count++;
		} else {
		    $template->{'pass'}=0;
		}
	    }
	    if($pass_count>1) {
		$stddev=sqrt($sq_sum)/($pass_count-1);
	    } else {
		$stddev=500;
	    }
	    $stddev=500 if($stddev<500);
 
	    $self->{'graph'}->{$node1}->{$node2}->{'stddev'}=$stddev;
	    #calculate the weight, favour abs small size
	    $max=$edges->{$node1}->{$node2}->{'count'};
	    $max=10 if($max>10);
	    my $weigth;
	    if($max>=2) {
		$weigth=$score_array[$max-1] - abs($average_gap);
	    } else {
		$weigth= - abs($average_gap);
	    }
	    $self->{'graph'}->{$node1}->{$node2}->{'weigth'}=$weigth;
	 #   print "for $node1 to $node2, gap is $average_gap with $max links, total link is $pass_count with std $stddev and weight of " . $self->{'graph'}->{$node1}->{$node2}->{'weigth'} . "\n" if ($max>=2);
	    #add the complement ones too to the graph
	    $self->{'graph'}->{$node2}->{$node1}{'ave_gap_size'}=$average_gap;
	    $self->{'graph'}->{$node2}->{$node1}{'count'}=$max;
	    $self->{'graph'}->{$node2}->{$node1}{'total_count'}=$#sorted+1;
	    $self->{'graph'}->{$node2}->{$node1}{'weigth'}=$weigth;	
	    $self->{'graph'}->{$node2}->{$node1}{'stddev'}=$stddev;
	    
	    $seen{$key}++;
	    $key=$node2 .'_'. $node1;
	    $seen{$key}++;
	}
    }
}
#------------------------------------------------
#add a precomputed edge to the final graph
#NOTE: it is important to add these edges last since
#the building internal edge process will reinitilize the
#edges
#------------------------------------------------
sub add_external_edge {
    my($project)=shift;
    
    my($node1_name, $node2_name, $weigth, $gap, $stddev)=@_;

    if(exists $project->{'graph'}{$node1_name}) {
	my $node1=$project->{'graph'}{$node1_name};
	$node1->{$node2_name} = {'ave_gap_size'=>$gap,
				 'stddev'=>$stddev,
				 'weigth'=>$weigth
				 }
    } else { # a new node1
	$project->{'graph'}{$node1_name} = {$node2_name=>{'ave_gap_size'=>$gap,
							  'stddev'=>$stddev,
							  'weigth'=>$weigth
							  }
					}
    }
}
#--------------------------------------------------------------------------------
#Purpose: add a new edge to the graph in the project, it is a directed edge
#Pre: a new edge has been prepared in a hash like:
#    {'node'=>[] #contig name with _left or _right indicating orientation
#     'gap_size'=>NUMBER,
#     'clone_contig_overlap'=>[],  #the overlapping size between clone and both contigs
#     'sample'=>ref(sample), #ref to the template that is the edge
#     'stddev'=>NUMBER,
#    }
#    the hash ref is passed into the sub.
#    The caller is the project
#Post: the new edge is added to the graph, two copies since it is undirected
#---------------------------------------------------------------------------------
sub add_edge_to_graph {

    my($project)=shift;
    my($edge)=shift;

    my($node1_name,  $node1_clone_arm, $node2_name, $node2_clone_arm);
    my($node1, $node2, $library_name);

    $node1_name=$edge->{'node'}[0];
    $node1_clone_arm=$edge->{'clone_contig_overlap'}[0];
    $node2_name=$edge->{'node'}[1];
    $node2_clone_arm=$edge->{'clone_contig_overlap'}[1];
    $library_name=$edge->{'library_name'};

    if(exists $project->{'graph'}{$node1_name}) {
	$node1=$project->{'graph'}{$node1_name};
        if(exists $node1->{$node2_name}) {
	    $node1->{$node2_name}{'total_gap'} += $edge->{'gap_size'};
	    $node1->{$node2_name}{'count'}++;
	    if(exists $node1->{$node2_name}{'samples'}{$library_name}) {
		$node1->{$node2_name}{'samples'}{$library_name}{'count'}++;
		if($node1->{$node2_name}{'samples'}{$library_name}{'max_left_arm'}<$node1_clone_arm) {
		    $node1->{$node2_name}{'samples'}{$library_name}{'max_left_arm'}=$node1_clone_arm;
		}
		if($node1->{$node2_name}{'samples'}{$library_name}{'max_rigth_arm'}<$node2_clone_arm) {
		    $node1->{$node2_name}{'samples'}{$library_name}{'max_rigth_arm'}=$node2_clone_arm;
		}
	    } else {  #it is a new library sample
		$node1->{$node2_name}{'samples'}{$library_name} = {'count'=>1,
								    'max_left_arm'=>$node1_clone_arm,
								    'max_rigth_arm'=>$node2_clone_arm,
								    'samples'=>[]
								}
	    }
	} else { #edge to link to a new node
	    $node1->{$node2_name} = {'total_gap'=>$edge->{'gap_size'},
				     'count'=>1,
				     'samples'=>{ $library_name => {'count'=>1,
								    'max_left_arm'=>$node1_clone_arm,
								    'max_rigth_arm'=>$node2_clone_arm,
								    'samples'=>[]
								}
					      }
				 }
	}
    } else { # a new node1
	$project->{'graph'}{$node1_name} = {$node2_name=>{'total_gap'=>$edge->{'gap_size'},
							  'count'=>1,
						          'samples'=>{ $library_name => {'count'=>1,
						       				     'max_left_arm'=>$node1_clone_arm,
						       				     'max_rigth_arm'=>$node2_clone_arm,
						       				     'samples'=>[]
						       				         }
								      }
						         }
					   }
    }

    push( @{$project->{'graph'}{$node1_name}{$node2_name}{'samples'}{$library_name}{'samples'}}, $edge->{'sample'} );

} #end of sub


#-------------------------------------------------------------------------------------------
#Purpose: Print out the graph of the project
#Format: NODE1\n
#        \t     Node1.1\t    Weight\t    Pairs
#              ... ...
#        NODE2\n
#        \t    Node2.1\t     Weight\t   Pairs
#              ... ...
#Pre: the project's graph has been built
#Post: Nothing changed, graph get printed out at STDOUT
#-------------------------------------------------------------------------------------------
sub print_graph {
    my($project)=shift;

    my($start_node, $end_node);

    foreach $start_node (sort keys %{$project->{'graph'}}) {
	#
	print "\n$start_node ->";
	foreach $end_node (sort keys %{$project->{'graph'}{$start_node}}) {
	    my $current_node=$project->{'graph'}{$start_node}{$end_node};
	    printf("\n\t%s\t%.2f\t%d\t%d", $end_node, 
		  $current_node->{'weigth'}, $current_node->{'count'}, $current_node->{'ave_gap_size'});
	}
	print "\n";
    }
}


########################################################
#print_graph_with_templates
#purpose: print out all edges with the template list that
#make up the edge
########################################################
sub print_graph_with_templates {
    my ($project)=shift;

    my($start_node, $end_node);

    foreach $start_node (sort keys %{$project->{'graph'}}) {
	#
	print "\n$start_node ->";
	foreach $end_node (sort keys %{$project->{'graph'}{$start_node}}) {
	    my $current_node=$project->{'graph'}{$start_node}{$end_node};
	    print "\n\t$end_node\n";
	    printf("\t\tScore: %.2f\%\n", $current_node->{'weigth'});
	    print "\t\tTotal Pairs: " . $current_node->{'count'} ."\n";
	    printf("\t\tGap Size: %d bp\n", $current_node->{'ave_gap_size'});
	    print "\t\tTemplate list:\n";
	    #print out the list
	    my $templatestr="";
	    my $count=0;
	    foreach my $library_name (keys %{$current_node->{'samples'}}) {
		my $sample_array_ref=$current_node->{'samples'}{$library_name}{'samples'};
		foreach my $sample (@$sample_array_ref) {
		    $templatestr .= $sample->[0]->mapgap_group() . ' ';
		    $count++;
		    if($count == 5) { #every five template perline
			$templatestr .= "\n\t\t";
			$count=0;
		    }
		}
	    }
	    print "\t\t$templatestr\n";
	}
	print "\n";
    }
}

###############################
#some utility subroutines
##############################
#------------------------------------------------------------------------------------
#incremental_find_path()
#Purpose: find the scaffold incrementally
#Algorithm:
#      get all the info, initially, each contig is a scaffold by its own
#      sort the edges based on their weight
#      while(edges whose count>=2) {
#           for the current edge(link between two scaffold node),
#              if(scaffold1 or scaffold2 is repetitive) {
#                      mark is as repetitive,
#                      next;
#              }
#              merge these two scaffold (caculate each contig position in the scaffold,
#                                        merge samples to the sample dump in the scaffold
#                                        )
#              update linkes involved in these two scaffolds(recaculate the gap size, bundle size and weigth)
#     }
#-----------------------------------------------------------------------------
sub incremental_find_path {

    my $self=shift;
    my $min_count=shift;
    my $flag=shift;  #if force, ignore the singlets
    
    $min_count=2 unless(defined $min_count);

    #get all the info needed for path building
    #the edges, the contigs and the current scaffolds
    my $edges=$self->{'graph'};
    my $contigs=$self->{'contig_scaffold'};
    my $scaffolds=$self->{'scaffold'};

    #sort the edges based on their weight, get the best one
 #   my ($node1, $node2) = &get_best_edge($edges, $min_count);
    
    #build a heap for all the eligible edges
    my $heap;
    $heap = new Heap::Fibonacci;

    #add all edges to the heap and get the highest one
    &build_heap($heap, $edges, $min_count);
    my ($node1, $node2)=&get_best_edge_from_heap($heap, $edges, $min_count);

    while($node1) { #there are edge fufill the criteria
	my $cur_edge=$edges->{$node1}{$node2};
	print STDERR "\ncheck edge between $node1 and $node2\n" if($VERBOSE);

	my @nodes_to_delete=();  #an array of repetitive nodes to be deleted
	if(my @temp=$self->is_repetitive($node1, $edges, $min_count, $flag)) {
	    push @nodes_to_delete, @temp;
	}
	if(my @temp=$self->is_repetitive($node2, $edges, $min_count, $flag)) {
	    push @nodes_to_delete, @temp;
	}
	#begin delete
	#record all the repetitive nodes so that in the future, all edges from these nodes will not be added
	foreach my $rep_node (@nodes_to_delete) {
	    
	    $self->{'rep_node'}{$rep_node}++;
	}

        #delete all edges from node1
	$self->delete_node(\@nodes_to_delete, $edges);
	
	#record all the repetitive nodes so that in the future, all edges from these nodes will not be added
	foreach my $rep_node (@nodes_to_delete) {
	    $self->{'rep_node'}{$rep_node}++;
	}

	#if any of them is repetitive, jump to the next edge
	if ($#nodes_to_delete>=0 || !defined $edges->{$node1}{$node2}{'ave_gap_size'}) {
	    #print "\ndelete $node1 or/and $node2\n" if($VERBOSE);
	    #($node1, $node2) = &get_best_edge($edges, $min_count);
	    ($node1, $node2)=&get_best_edge_from_heap($heap, $edges, $min_count);
	    next;
	}

	#if the gap size is too small, with -20k as the cutoff
	if( ($cur_edge->{'ave_gap_size'} < -20000) && (abs($cur_edge->{'ave_gap_size'})>1.5*$cur_edge->{'stddev'})) {
	    $self->delete_edge($node1, $node2);
	    #($node1, $node2) = &get_best_edge($edges, $min_count);
	    ($node1, $node2)=&get_best_edge_from_heap($heap, $edges, $min_count);
	    next;
	}

	#now here is a good edge to use, merge two scaffolds and update the edges
	print STDERR "\nBegin to merge $node1 and $node2\n" if($VERBOSE);
	print STDERR "The gap size is ".$edges->{$node1}{$node2}{'ave_gap_size'}. " with " .$edges->{$node1}{$node2}{'count'} . " links\n" if($VERBOSE); 
	my $new_scaffold=$self->merge_scaffolds($node1, $node2, $heap);
	
	$self->print_scaffold($new_scaffold) if($VERBOSE);
	
	#get a new round of merging
	($node1, $node2)=&get_best_edge_from_heap($heap, $edges, $min_count);
	#($node1, $node2) = &get_best_edge($edges, $min_count);
    }

    
}


sub build_heap {

    my ($heap, $edges, $min_count)=@_;

    foreach my $node1 (keys %{$edges}) {
	foreach my $node2 (keys %{$edges->{$node1}}) {
	    if (exists $edges->{$node1}->{$node2}{'count'} && $edges->{$node1}->{$node2}{'count'}>=$min_count) {
		my $ele=Atlas::ScaffoldHeapEle->new($edges->{$node1}->{$node2}{'weigth'}, $node1, $node2);
		$heap->add($ele);
	    }
	}
    }

}

sub get_best_edge_from_heap {

    my($heap, $edges, $min_count)=@_;
    
    $min_count=2 unless defined $min_count;
    
    my $ele=$heap->extract_minimum;
    while($ele) {
	#check the edge if good or not
	my $node1=$ele->{'node1'};
	my $node2=$ele->{'node2'};
#	print STDOUT $edges->{$node1}{$node2}{'count'} ."\t" . $edges->{$node1}->{$node2}{'weigth'} . "\t" . $ele->{'val'} ."\n";
	last if(exists $edges->{$node1} && exists $edges->{$node1}{$node2} && exists $edges->{$node1}{$node2}{'count'} && $edges->{$node1}->{$node2}{'count'}>=$min_count && $edges->{$node1}->{$node2}{'weigth'}==$ele->{'val'}); #the same edge betweent heap and graph

	$ele=$heap->extract_minimum;
    }

    if(defined $ele) {
	return ($ele->{'node1'}, $ele->{'node2'});
    }

    return;
}


sub add_edge_to_heap {

    my $self=shift;
    my ($heap, $in_node1, $in_node2)=@_;

    my $edges;
    if($in_node1=~/[_left|_right]$/ && $in_node2=~/[_left|_right]$/) {
	my $temp=$self->{'graph'}->{$in_node1}->{$in_node2};
	$edges->{$in_node1}={$in_node2=>$temp};
    } else {
	$edges->{$in_node1.'_left'}{$in_node2.'_left'}=$self->{'graph'}{$in_node1.'_left'}{$in_node2.'_left'}
	       if(exists $self->{'graph'}{$in_node1.'_left'}{$in_node2.'_left'});
	$edges->{$in_node1.'_left'}{$in_node2.'_right'}=$self->{'graph'}{$in_node1.'_left'}{$in_node2.'_right'}
	       if(exists $self->{'graph'}{$in_node1.'_left'}{$in_node2.'_right'});
	$edges->{$in_node1.'_right'}{$in_node2.'_left'}=$self->{'graph'}{$in_node1.'_right'}{$in_node2.'_left'}
	       if(exists $self->{'graph'}{$in_node1.'_right'}{$in_node2.'_left'});
	$edges->{$in_node1.'_right'}{$in_node2.'_right'}=$self->{'graph'}{$in_node1.'_right'}{$in_node2.'_right'}
	       if(exists $self->{'graph'}{$in_node1.'_right'}{$in_node2.'_right'});
    }
   
    #get the edge out then add to the heap
    foreach  my $node1 (keys %{$edges}) {
	foreach my $node2 (keys %{$edges->{$node1}}) {
	    if (exists $edges->{$node1}->{$node2}{'weigth'}) {
		my $ele=Atlas::ScaffoldHeapEle->new($edges->{$node1}->{$node2}{'weigth'}, $node1, $node2);
		$heap->add($ele);
	    }
	}
    }
}

#clean up all the repetitive nodes before build the scaffold
#effect unknown

sub clean_node {

    my $self=shift;
    my $min_count=shift;

    $min_count=2 unless(defined $min_count);

    #get all the info needed for path building
    #the edges, the contigs and the current scaffolds
    my $edges=$self->{'graph'};
    my $contigs=$self->{'contig_scaffold'};
    my $scaffolds=$self->{'scaffold'};
        
    #clean up repetitive nodes
    my @single_nodes_to_delete=();
    foreach my $contig_name (keys %{$scaffolds}) {
	print "test $contig_name\n";
	
	my $node=$contig_name . '_left';
	if(my @temp=$self->single_node_is_repetitive($node, $edges, $min_count)) {
	    push @single_nodes_to_delete, @temp;
	}
	$node=$contig_name . '_right';
	if(my @temp=$self->single_node_is_repetitive($node, $edges, $min_count)) {
	    push @single_nodes_to_delete, @temp;
	}
	#record all the repetitive nodes so that in the future, all edges from these nodes will not be added
    }

    foreach my $rep_node (@single_nodes_to_delete) {    
	$self->{'rep_node'}{$rep_node}++;
    }

    #delete all edges from node1
    $self->delete_node(\@single_nodes_to_delete, $edges);
	
    #record all the repetitive nodes so that in the future, all edges from these nodes will not be added
    foreach my $rep_node (@single_nodes_to_delete) {
	$self->{'rep_node'}{$rep_node}++;
    }
}

#--------------------
=head1 find_split_node

    to find all the misjoined nodes and give the potential split positions


    $self->find_split_node();

=cut
sub find_split_node {

    my $self=shift;
    my $min_count=shift;

    $min_count=2 unless(defined $min_count);

    #get all the info needed for path building
    #the edges, the contigs and the current scaffolds
    my $edges=$self->{'graph'};
    my $contigs=$self->{'contig_scaffold'};
    my $scaffolds=$self->{'scaffold'};
        
    #clean up repetitive nodes
    my @split_node=();
    foreach my $contig_name (keys %{$scaffolds}) {
	#print "test $contig_name\n";
	
	if(my @temp=$self->node_is_misjoin($contig_name, $edges, $min_count)) {
	    push @split_node, @temp;
	}
    }

    foreach my $node (@split_node) {
	print $node->{'node'} . "\t" . $node->{'side'} . "\t" . $node->{'start'} . "\t" . $node->{'end'} . "\n";
    }
}

sub node_is_misjoin {

    my $self=shift;
    my($node, $edges, $min_count)=@_;

    #check both the left side and right side of the node
    my @side=('left', 'right');
    my @splits=();
    
    #get the window size that call a potential split
    my $window=&set_bundle_size();

    foreach my $side (@side) {
	my $cur_node=$node . '_' . $side;
	my @next_nodes=keys %{$edges->{$cur_node}};
	foreach my $next_node (@next_nodes) {
	    #if this edge only has one link, ignore
	    next if $edges->{$cur_node}{$next_node}{'count'}<$min_count;
	   # print "check edge between $cur_node and $next_node\n";

	    #check gap size, if the gap size is less than -2*window, potential split
	    if($edges->{$cur_node}{$next_node}{'ave_gap_size'}+$window*2<0) {
		#get each node size and all the samples that contribute the link, find their footprint
		my @template=@{$edges->{$cur_node}->{$next_node}{'sample'}};
		my @sample=();
		foreach my $template (@template) {
		    push @sample, $template->{'sample'} if($template->{'pass'}==1);
		}
		
		#foreach sample, find the trace fall in the node/contig
		my @trace=();
		foreach my $sample_ref (@sample) {
		    foreach my $trace (@{$sample_ref}) {
			push @trace, $trace if ($trace->contig()->name() eq $node);
		    }
		}

		#for each trace in the node, based on the side, find the right/left most trace and find
		#the average right/left boundary
		my ($start, $end, $span);
		#sort all the trace by their contig start
		@trace=sort { $a->contig_start() <=> $b->contig_start() } @trace;
		$start=$trace[0]->contig_start();
		$end=$trace[-1]->contig_end();
		my $contig_start=$trace[0]->contig()->start();
		my $contig_end=$trace[0]->contig()->end();
		if($side eq 'left') {
		    foreach my $trace (@trace) {
			my $library_name=$trace->coslib();
			my $library=$self->{'contig_pair_samples'}{$library_name};
			$span += $trace->contig_start()-$library->{'average_size'};
		    }
		} else {
		    foreach my $trace (@trace) {
			my $library_name=$trace->coslib();
			my $library=$self->{'contig_pair_samples'}{$library_name};
			$span += $trace->contig_end()+$library->{'average_size'};
		    }
		}
		
		my $ave_span= int($span/($#trace+1));
		
		#save the data
		if($side eq 'left') {
		    next if ($ave_span < $contig_start && $start < ($contig_start+2000) );  #it is really a tail
		    push @splits, {'node'=>$node,
				   'side'=>$side,
				   'start'=>$ave_span,
				   'end'=>$start};
		} else {
		    next if ($ave_span > $contig_end  && $end > ($contig_end-2000) ); #a tail
		    push @splits, {'node'=>$node,
				   'side'=>$side,
				   'start'=>$end,
				   'end'=>$ave_span};
		}
	    }
	}
    }

    return @splits;

}
#-----------------------------------------------
#gap_fill
#fill up the gaps in a scaffold by looking for cases that one singleton 
#is linked to adjacent contigs in a scaffold and can be fit in
#Only edges left after incremental building of the scaffold are used
#However, due to the inprecision of one link, the repetitiveness is not check since
#it is no point
#------------------------------------------------------------------
sub gap_fill {

    my $self=shift;

    my $edges=$self->{'graph'};
    my $scaffold=$self->{'scaffold'};
    my $contigs=$self->{'contig_scaffold'};
    my $window=&set_bundle_size();

    #find the list of singlets that linked to the same scaffold both side
    foreach my $node1 (keys %{$edges}) {
	$node1=~/^(\S+)_(\S+)$/;
	my $contig1=$1;
	next if($1=~/_/); #skip unless it is a singlet
	my $com_node1;
	if($2 eq 'left') {
	    $com_node1=$1 . '_right';
	} else {
	    $com_node1=$1 .'_left';
	}
	next unless(exists $edges->{$com_node1});

	print "check $node1\n" if($VERBOSE);
	my $count=0;
	my $com_node2;
	my $node2; #the scaffold the contig try to fit in
	my $sca;
	foreach my $node22 (keys %{$edges->{$node1}}) {
	    $node22=~/^(\S+)_(\S+)$/;
	    my $scaid=$1;
	    my $tail=$2;
	    next unless ($scaid=~/_/); #only count if the second node is a scaffold
	    my $com_node22;
	    if($tail eq 'left') {
		$com_node22=$scaid . '_right';
	    } else {
		$com_node22=$scaid .'_left';
	    }
	    next unless ( exists $edges->{$com_node1}{$com_node22});
	    $count++;
	    $sca=$scaid;
	    $node2=$node22;
	    $com_node2=$com_node22;
	}

	if($count==1) {
	    #fit it in if possible
	    print "Fit $node1 into scaffold $sca\n" if($VERBOSE);
	    #check if the position estimated by both end is accurate enough
	    my ($putative_start1, $putative_start2);
	    if($node2=~/_right/) {
		$putative_start1=$scaffold->{$sca}{'size'}+$edges->{$node1}{$node2}{'ave_gap_size'};
		$putative_start2=-$edges->{$com_node1}{$com_node2}{'ave_gap_size'}-$scaffold->{$contig1}{'size'};
	    } else {
		$putative_start1=-$edges->{$node1}{$node2}{'ave_gap_size'}-$scaffold->{$contig1}{'size'};
		$putative_start2=$scaffold->{$sca}{'size'}+$edges->{$com_node1}{$com_node2}{'ave_gap_size'};
	    }

	    next unless (abs($putative_start1 - $putative_start2)<$window);

	    #find the two adjancent contigs that the contig suppose to go
	    #look at the sample for all the potential contig
	    my %sca_contig=();
	    my @contigs=();
	    foreach my $contig_name (keys %{$scaffold->{$sca}}) {
		push @contigs, $contig_name 
		    unless ($contig_name eq 'size' || $contig_name eq 'head' 
			    || $contig_name eq 'tail' || $contig_name eq 'sample'); 
	    
	    }

	    @contigs=sort {$scaffold->{$sca}{$a}{'start'} <=> $scaffold->{$sca}{$b}{'start'}}
	                     @contigs; 
	    for(my $i=0; $i<=$#contigs; $i++) {
		my $cur_contig=$scaffold->{$sca}{$contigs[$i]};
		$sca_contig{$contigs[$i]}={'pos'=>$i,
				      'start'=>$cur_contig->{'start'},
				      'end'=> $cur_contig->{'end'}};
		$sca_contig{$contigs[$i]}->{'next'}=$contigs[$i+1] if($i<$#contigs);
	    }

	    #find the left side contigs
	    my %left_contig=();
	    my %right_contig=();
	    my $orientation;
	    if($node2=~/_right/) {
		if($node1=~/_right/) {
		    $orientation='-';
		} else {
		    $orientation='+';
		}
		#get all the samples
		foreach my $template (@{$edges->{$node1}{$node2}{'sample'}}) {
		    my $sample=$template->{'sample'};
			#get the trace
			foreach my $trace (@{$sample}) {
			    my $contig_name=$trace->contig()->name();
			    next unless exists $sca_contig{$contig_name};
			    next unless(abs($sca_contig{$contig_name}->{'end'} - $putative_start1)<$window);
			    $left_contig{$contig_name}++;
			}
		    
		}
		foreach my $template (@{$edges->{$com_node1}{$com_node2}{'sample'}}) {
		    my $sample=$template->{'sample'}; 
			#get the trace
			foreach my $trace (@{$sample}) {
			    my $contig_name=$trace->contig()->name();
			    next unless exists $sca_contig{$contig_name};
			    next unless(abs($sca_contig{$contig_name}->{'start'} - $putative_start2)<$window);
			    $right_contig{$contig_name}++;
			}
		    
		}
	    } else {
		if($node1=~/_right/) {
		    $orientation='+';
		} else {
		    $orientation='-';
		}
		#get all the samples
		foreach my $template (@{$edges->{$com_node1}{$com_node2}{'sample'}}) {
		     my $sample=$template->{'sample'};
			#get the trace
			foreach my $trace (@{$sample}) {
			    my $contig_name=$trace->contig()->name();
			    next unless exists $sca_contig{$contig_name};
			    next unless(abs($sca_contig{$contig_name}->{'end'} - $putative_start1)<$window);
			    $left_contig{$contig_name}++;
			}
		    
		}
		foreach my $template (@{$edges->{$node1}{$node2}{'sample'}}) {
		    my $sample=$template->{'sample'};
			#get the trace
			foreach my $trace (@{$sample}) {
			    my $contig_name=$trace->contig()->name();
			    next unless exists $sca_contig{$contig_name};
			    next unless(abs($sca_contig{$contig_name}->{'start'} - $putative_start2)<$window);
			    $right_contig{$contig_name}++;
			}
		
		}
	    }

	    my $pair_count=0;
	    my ($left, $right);
	    foreach my $left_contig (keys %left_contig) {
		my $right_contig=$sca_contig{$left_contig}->{'next'};
		next unless defined $right_contig;
		next unless exists $right_contig{$right_contig};
		my $gap=$sca_contig{$right_contig}->{'start'} - $sca_contig{$left_contig}->{'end'};
		next if($gap<$scaffold->{$contig1}->{'size'}-4000);
		$pair_count++;
		$left=$left_contig;
		$right=$right_contig;
	    }

	    next unless $pair_count==1;
	    my $putative_pos = ( $putative_start1 +  $putative_start2)/2;
	    if($putative_pos >= $sca_contig{$left}->{'end'} 
	       && $putative_pos <= $sca_contig{$right}->{'start'}) {
		#fit right in
		$scaffold->{$sca}->{$contig1}={'start'=>$putative_pos,
					       'end'=>$putative_pos+$scaffold->{$contig1}->{'size'}-1,
					       'orientation'=>$orientation};
	    } else { #place in the middle
		my $offset= ($sca_contig{$right}->{'start'} - $sca_contig{$left}->{'end'}
		              - $scaffold->{$contig1}->{'size'})/2;
		$scaffold->{$sca}->{$contig1}={'start'=>$sca_contig{$left}->{'end'}+$offset,
					       'end'=>$sca_contig{$right}->{'start'}-$offset,
					       'orientation'=>$orientation};
	    }
	    
	    print "FITS IN\n" if($VERBOSE);
	    #delete all the edges involved in contig1
	    my @old_nodes=($node1, $com_node1);
	    foreach my $old_node (@old_nodes) {
		foreach my $old_node2 ( keys %{$edges->{$old_node}}) {
		    delete $edges->{$old_node2}{$old_node};
		}
		delete $edges->{$old_node};
	    }

	    #change the scaffold id
	    my $old_scaid=$contigs->{$contig1};
	   # print "delete $old_scaid\n";
	    delete $scaffold->{$old_scaid};
	    $contigs->{$contig1}=$sca;
	} #end of fitting
    }
}




#-------------------------
#print scaffold
#---------------------------
sub print_scaffold {
    my $self=shift;
    my $scaffold_id=shift;

    my $scaffolds=$self->{'scaffold'};

    if(defined $scaffold_id) {
	print "$scaffold_id\n";
	printf("%d\n", $scaffolds->{$scaffold_id}{'size'});
	my @contigs=();
	foreach my $contig_name (keys %{$scaffolds->{$scaffold_id}}) {
	    push @contigs, $contig_name 
		    unless ($contig_name eq 'size' || $contig_name eq 'head' || $contig_name eq 'tail' || $contig_name eq 'sample'); 
	    
	}
	foreach my $contig (@contigs) {
	    next  if($contig eq 'size' || $contig eq 'head' || $contig eq 'tail' || $contig eq 'sample');
	    my $cur_contig=$scaffolds->{$scaffold_id}{$contig};
	    print "$contig:\t";
	    printf("%d\t%d\t%s\n", $cur_contig->{'start'}, $cur_contig->{'end'}, $cur_contig->{'orientation'});
	}
    } else {
	foreach my $scaffold (keys %{$scaffolds}) {
	    print "SCAFFOLD:\t$scaffold\t";
	    printf("%d\n", $scaffolds->{$scaffold}{'size'});
	    #sort all the contigs in the scaffold and print out
	    my @contigs=();
	    foreach my $contig_name (keys %{$scaffolds->{$scaffold}}) {
		push @contigs, $contig_name 
		    unless ($contig_name eq 'size' || $contig_name eq 'head' || $contig_name eq 'tail' || $contig_name eq 'sample'); 
	    
		}

	    @contigs=sort {$scaffolds->{$scaffold}{$a}{'start'} <=> $scaffolds->{$scaffold}{$b}{'start'}}
	                     @contigs; 
	    foreach my $contig (@contigs) {
		next  if($contig eq 'size' || $contig eq 'head' || $contig eq 'tail' || $contig eq 'sample');
		my $cur_contig=$scaffolds->{$scaffold}{$contig};
		print "$contig:\t";
		printf("%d\t%d\t%s\n", $cur_contig->{'start'}, $cur_contig->{'end'}, $cur_contig->{'orientation'});
	    }
	}
    }
}
#------------------------------------------------------
#merge_scaffolds
#to merge two connected scaffolds
#algorithm:
#     get the edge that connects these two scaffolds
#     based on the edge, get the head and tail of the new scaffold and get the new id
#     (reverse scaffod1 and scaffold2 if necessary)
#     loop through scaffold1 and 2 contigs, update there scaffold lookup and recalculate the pos
#     delete the old entry for scaffold1 and scaffold2 in the self->{'scaffold'}
#     push all the samples that connect the old scaffold1 or 2 to other scaffolds into a new array
#     At the meantime, delete all the entries involved in scaffold1 or 2 int the self->{'path'}
#     recalculate all the samples in the array and add them to the path
#------------------------------------------------------------------------
sub merge_scaffolds {

    my $self=shift;
    my ($node1, $node2, $heap)=@_;

    my $edges=$self->{'graph'};
    my $scaffold=$self->{'scaffold'};
    my $contigs=$self->{'contig_scaffold'};

    my $cur_edge=$edges->{$node1}{$node2};
    unless (exists $cur_edge->{'sample'}) {
	my $temp=$node1;
	$node1=$node2;
	$node2=$temp;
	$cur_edge=$edges->{$node1}{$node2};
    } #get the one has sample attached
    
    $node1=~/^(\S+)_(\S+)$/;
    my $scaffold1=$1;
    my $side1=$2;
    $node2=~/^(\S+)_(\S+)$/;
    my $scaffold2=$1;
    my $side2=$2;
    #determine the id for the new scaffold
    my ($head, $tail);
    if($side1 eq 'left') { #reverse
	$head=$scaffold->{$scaffold1}->{'tail'};
    } else {
	$head=$scaffold->{$scaffold1}->{'head'};
    }
 
    if($side2 eq 'right') { #reverse
	$tail=$scaffold->{$scaffold2}->{'head'};
    } else {
	$tail=$scaffold->{$scaffold2}->{'tail'};
    }
    my $scaffold_id=$head .'_'.$tail;

    my $scaffold_size1=$scaffold->{$scaffold1}->{'size'};
    if($side1 eq 'left') { #need to reverse the scaffold
	foreach my $contig (keys %{$scaffold->{$scaffold1}}) {
	    next if($contig eq 'size' || $contig eq 'head' || $contig eq 'tail' || $contig eq 'sample');
	    $scaffold->{$scaffold_id}{$contig}{'end'}=$scaffold_size1-$scaffold->{$scaffold1}{$contig}{'start'}+1;
	    $scaffold->{$scaffold_id}{$contig}{'start'}=$scaffold_size1-$scaffold->{$scaffold1}{$contig}{'end'}+1;
	    if($scaffold->{$scaffold1}{$contig}{'orientation'} eq '-') {
		$scaffold->{$scaffold_id}{$contig}{'orientation'}='+';
	    } else {
		$scaffold->{$scaffold_id}{$contig}{'orientation'}='-';
	    }
	    $contigs->{$contig}=$scaffold_id; #update the lookup table
	}
    } else { # noththing changed except the contig lookup table
	foreach my $contig (keys %{$scaffold->{$scaffold1}}) {
	    next if($contig eq 'size' || $contig eq 'head' || $contig eq 'tail' || $contig eq 'sample');
	    $contigs->{$contig}=$scaffold_id; #update the lookup table
	    $scaffold->{$scaffold_id}{$contig}{'start'}=$scaffold->{$scaffold1}{$contig}{'start'};
	    $scaffold->{$scaffold_id}{$contig}{'end'}=$scaffold->{$scaffold1}{$contig}{'end'};
	    $scaffold->{$scaffold_id}{$contig}{'orientation'}=$scaffold->{$scaffold1}{$contig}{'orientation'};
	    $contigs->{$contig}=$scaffold_id; #update the lookup table 
	}
    }
     
    my $scaffold_size2=$scaffold->{$scaffold2}->{'size'};
    my $gap_size=$cur_edge->{'ave_gap_size'};
    $gap_size=500 if ($gap_size<-5000);
    if($side2 eq 'right') { #need to reverse the scaffold
	foreach my $contig (keys %{$scaffold->{$scaffold2}}) {
	    next if($contig eq 'size' || $contig eq 'head' || $contig eq 'tail' || $contig eq 'sample');
	    $scaffold->{$scaffold_id}{$contig}{'end'}=$scaffold_size1+$gap_size+$scaffold_size2-$scaffold->{$scaffold2}{$contig}{'start'}+1;
	    $scaffold->{$scaffold_id}{$contig}{'start'}=$scaffold_size1+$gap_size+$scaffold_size2-$scaffold->{$scaffold2}{$contig}{'end'}+1;
	    if($scaffold->{$scaffold2}{$contig}{'orientation'} eq '-') {
		$scaffold->{$scaffold_id}{$contig}{'orientation'}='+';
	    } else {
		$scaffold->{$scaffold_id}{$contig}{'orientation'}='-';
	    }
	    $contigs->{$contig}=$scaffold_id; #update the lookup table
	}
    } else { #no reverse necessary
	foreach my $contig (keys %{$scaffold->{$scaffold2}}) {
	    next if($contig eq 'size' || $contig eq 'head' || $contig eq 'tail' || $contig eq 'sample');
	    $scaffold->{$scaffold_id}{$contig}{'start'}=$scaffold_size1+$gap_size+$scaffold->{$scaffold2}{$contig}{'start'};
	    $scaffold->{$scaffold_id}{$contig}{'end'}=$scaffold_size1+$gap_size+$scaffold->{$scaffold2}{$contig}{'end'};
	    $scaffold->{$scaffold_id}{$contig}{'orientation'}=$scaffold->{$scaffold2}{$contig}{'orientation'};
	    $contigs->{$contig}=$scaffold_id; #update the lookup table
	}
    }

    $scaffold->{$scaffold_id}{'head'}=$head;
    $scaffold->{$scaffold_id}{'tail'}=$tail;
    $scaffold->{$scaffold_id}{'size'}=$scaffold_size1+$gap_size+$scaffold_size2;
    #push all the samples from the cur_edge into the new scaffold
    foreach my $sample (@{$cur_edge->{'sample'}}) {
	push @{$scaffold->{$scaffold_id}{'sample'}}, $sample->{'sample'};
    }
    #delete the cur_edge
    delete $edges->{$node1}{$node2};
    delete $edges->{$node2}{$node1};

    #get all the old scaffolds sample
    push @{$scaffold->{$scaffold_id}{'sample'}}, @{$scaffold->{$scaffold1}{'sample'}} if (exists $scaffold->{$scaffold1}{'sample'});
    push @{$scaffold->{$scaffold_id}{'sample'}}, @{$scaffold->{$scaffold2}{'sample'}} if (exists $scaffold->{$scaffold2}{'sample'});

    #now delete the old scaffold entries
    delete $scaffold->{$scaffold1};
    delete $scaffold->{$scaffold2};
   # print "delete scaffold $scaffold1 and $scaffold2\n";
    
    #add new entry to contigs to scaffoldid
    $contigs->{$scaffold1}=$scaffold_id;
    $contigs->{$scaffold2}=$scaffold_id;
    $contigs->{$scaffold_id}=$scaffold_id;

    #update the edges
    #find all the templates that connect the old scaffolds to other stuff
    #get all the four nodes
    my @old_nodes=(($scaffold1 . '_left'), ($scaffold1 . '_right'),  ($scaffold2 . '_left'), ($scaffold2 . '_right') );
    #get all other nodes connected to any of these four nodes
    my @samples=();
    my %touched_node=();
    foreach my $old_node (@old_nodes) {
	next unless exists $edges->{$old_node};
	foreach my $next_old_node (keys %{$edges->{$old_node}}) {
	    foreach my $old_sample (@{$edges->{$next_old_node}{$old_node}->{'sample'}}) {
		push @samples, $old_sample->{'sample'};
	    }
	    #delete the old edge
	    delete $edges->{$old_node}{$next_old_node};
	    delete $edges->{$next_old_node}{$old_node};
	    #my @key= keys %{$edges->{$next_old_node}};
	    #delete $edges->{$next_old_node} unless ($#key>=0);
	    
	    $old_node=~/^(\S+)_(\S+)$/;
	    my $t_node1=$1;
	    $next_old_node=~/^(\S+)_(\S+)$/;
	    my $t_node2=$1;
	#    print "$t_node1 and t_node2 is $t_node2\n";
	    $touched_node{$contigs->{$t_node1}}->{$contigs->{$t_node2}}++;
	}
	delete $edges->{$old_node};
    }

    #recalculate the new set of samples and add the new edges
    $self->add_sample_to_graph(\@samples);
    foreach my $touch_node1 (keys %touched_node) {
	foreach my $touch_node2 (keys %{$touched_node{$touch_node1}}) {
	    $self->Cal_Bundle($touch_node1, $touch_node2);
	    $self->add_edge_to_heap($heap, $touch_node1, $touch_node2);
	}
    }
    return $scaffold_id;
}

#
sub delete_edge {

    my $self=shift;
    my ($node1, $node2)=@_;

    ($node1, $node2)=sort($node1, $node2);
    my $edges=$self->{'graph'};

    foreach my $template (@{$edges->{$node1}{$node2}{'sample'}}) {
	next unless $template->{'pass'}==1;
	#get the contig name
	my $sample=$template->{'sample'};
	
	push @{$self->{'delete_sample'}}, $sample;  #record it	
    }
    
    delete $edges->{$node1}{$node2};
    delete $edges->{$node2}{$node1};
}

#-----------------------------------------------
#delete all the edges come from a particular node
#
#----------------------------------------------
sub delete_node {

    my $self=shift;
    my ($nodes, $edges)=@_;

    my $rep_node=$self->{'rep_node'};
    my $contigs=$self->{'contig_scaffold'};

    #note that both direction of a edge are stored in the graph
    my %potential_rep_node=();
    foreach my $cur_node (@{$nodes}) {
	print "delete $cur_node\n" if ($VERBOSE);
	next unless exists $edges->{$cur_node};
	$cur_node=~/(\S+)_(\S+)$/;
	my $cur_scaffold=$1;
	foreach my $next_node (keys %{$edges->{$cur_node}}) {
	    #find out the contigs involved in these links
	    next unless (defined $edges->{$cur_node}{$next_node}{'count'});
	    if($edges->{$cur_node}{$next_node}{'count'}>=2) { #only when it is not a single count, go ahead see
		#get every smaples that fall in the window and record the contig name
		foreach my $template (@{$edges->{$cur_node}{$next_node}{'sample'}}) {
		    my $sample=$template->{'sample'};
		    push @{$self->{'delete_sample'}}, $sample;  #record it
		    next unless $template->{'pass'}==1;

		    #get the contig name if it appears more than once
		    for(my $i=0; $i <= $#$sample; $i++) {
			my $contig=$sample->[$i]->contig();
			my $contig_name=$contig->name();
			my $strand=$sample->[$i]->contig_strand();
			next unless($contigs->{$contig_name} eq $cur_scaffold); 
			#begin to record clone end
			my $clone_end;
			if ($strand eq '+') {
			    $clone_end = $contig_name . '_right';
			} else {
			    $clone_end = $contig_name . '_left';
			}
			$potential_rep_node{$clone_end}++;
		    } #end of for loop
		}
	    }    
	    delete $edges->{$next_node}{$cur_node};
	    delete $edges->{$cur_node}{$next_node};
	}
    }
    
    foreach my $end (keys %potential_rep_node) {
	$rep_node->{$end}++ if($potential_rep_node{$end}>1);
    }

}
#---------------------------------------------------------
#test if the node is a repetitive node
#return value is the list of repetitive nodes
#--------------------------------------------------------
sub is_repetitive {
    my $self=shift;
    my($node, $edges, $min_count, $force)=@_;

    $min_count=2 unless (defined $min_count);

    my $scaffold=$self->{'scaffold'};
    my $contigs=$self->{'contig_scaffold'};

    #in the edges, it is a undirected graph with edges of both diretion stored
    my %repetitive_node=();
    #sort all the edges connected to node1 based on the gap size(distance from the node)
    my @node_keys=();
    foreach my $node_key (keys %{$edges->{$node}}) {
	push @node_keys, $node_key if(exists $edges->{$node}{$node_key}->{'ave_gap_size'});
    }
    my @next_nodes=sort {$edges->{$node}{$a}->{'ave_gap_size'} <=> $edges->{$node}{$b}->{'ave_gap_size'}}
                       @node_keys;

    my $end=0;
#    print "test on node $node\n";
    foreach my $next_node (@next_nodes) {
	next unless (defined $edges->{$node}{$next_node}{'count'});
	#if this edge only has one link, ignore
	next if $edges->{$node}{$next_node}{'count'}<$min_count;
	$next_node=~/^(\S+)_(\S+)$/;
	#print "next node is $next_node and $1\n" if($VERBOSE);
	next if (defined $force && !($1=~/_/) ); #signlet ignored if desired
 
#	print "next node is $next_node under testing\n";
	print "$node and $next_node gap size ". $edges->{$node}{$next_node}{'ave_gap_size'} . " with stddev of " . $edges->{$node}{$next_node}{'stddev'} ."\n" if($VERBOSE);

	#check gap size, both node repetitive if it is too negative
	if($edges->{$node}{$next_node}{'ave_gap_size'}<(-2000-4*$edges->{$node}{$next_node}{'stddev'})) {
	    #in case of overlapping, if the link count is 2, just delete the link instead of both nodes
	    if($edges->{$node}{$next_node}{'count'}==2) {
		$self->delete_edge($node, $next_node);
		print "Delete edges only for overlapping $node and $next_node\n" if($VERBOSE);
		next;
	    } else {
		$repetitive_node{$node}++;
		$repetitive_node{$next_node}++;
		print "Overlapping $node and $next_node\n" if($VERBOSE);
	    }
	}

	#check if nodes linked to the node tested is on top of each other
	if( ($edges->{$node}{$next_node}{'ave_gap_size'}-$end)<(-2000-4*$edges->{$node}{$next_node}{'stddev'})) {
	    $repetitive_node{$node}++;
	    print "Fork node $node\n" if($VERBOSE);
	}
	
	$next_node=~/^(\S+)_(\S+)$/;
	$end=$edges->{$node}{$next_node}{'ave_gap_size'}+$scaffold->{$1}{'size'};
#	print "end is $end\n";
	
    }

    return (keys %repetitive_node);
}

#---------------------------------------------------------
#test if the node is a repetitive node
#return value is the list of repetitive nodes
#--------------------------------------------------------
sub single_node_is_repetitive {
    my $self=shift;
    my($node, $edges, $min_count)=@_;

    $min_count=2 unless (defined $min_count);

    my $scaffold=$self->{'scaffold'};

    #in the edges, it is a undirected graph with edges of both diretion stored
    my %repetitive_node=();
    #sort all the edges connected to node1 based on the gap size(distance from the node)
    my @next_nodes=sort {$edges->{$node}{$a}->{'ave_gap_size'} <=> $edges->{$node}{$b}->{'ave_gap_size'}}
                        (keys %{$edges->{$node}});

    my $end=0;
#    print "test on node $node\n";
    foreach my $next_node (@next_nodes) {
	#if this edge only has one link, ignore
	next if $edges->{$node}{$next_node}{'count'}<$min_count;
	print "check edge between $node and $next_node\n";
	print "stddev is " . $edges->{$node}{$next_node}{'stddev'} ."\n" if($VERBOSE);

	#check gap size, both node repetitive if it is too negative
	if($edges->{$node}{$next_node}{'ave_gap_size'}<(-2000-4*$edges->{$node}{$next_node}{'stddev'})) {
	    $repetitive_node{$node}++;
	    print "Overlapping $node and $next_node\n" if($VERBOSE);
	}

	#check if nodes linked to the node tested is on top of each other
	if( ($edges->{$node}{$next_node}{'ave_gap_size'}-$end)<(-2000-4*$edges->{$node}{$next_node}{'stddev'})) {
	    $repetitive_node{$node}++;
	    print "Fork node $node\n" if($VERBOSE);
	}
	
	$next_node=~/^(\S+)_(\S+)$/;
	$end=$edges->{$node}{$next_node}{'ave_gap_size'}+$scaffold->{$1}{'size'};
#	print "end is $end\n";
	
    }

    return (keys %repetitive_node);
}

#----------------------------------------------------------
#
#-----------------------------------------------------
sub get_best_edge {

    my $edges=shift;
    my $min_count=shift;
    
    $min_count=2 unless (defined $min_count);

    #sort all the edges based on their weight, in the future, use priotity queue
    my @edges=();
    
    foreach my $node1 (keys %{$edges}) {
	foreach my $node2 (keys %{$edges->{$node1}}) {
	    push @edges, {'node1'=>$node1,
			  'node2'=>$node2,
			  'weigth'=>$edges->{$node1}->{$node2}{'weigth'}}
		    if (defined $edges->{$node1}->{$node2}{'count'} && $edges->{$node1}->{$node2}{'count'}>=$min_count);
	    #print "$node1, $node2 with weight of" . $edges->{$node1}->{$node2}{'weigth'} ."\n";
	}
    }

    return if($#edges<0);

    my @sort_edges=sort {$b->{'weigth'} <=> $a->{'weigth'}} @edges;

    my ($node1, $node2) = sort ($sort_edges[0]->{'node1'}, $sort_edges[0]->{'node2'});

    return ($node1, $node2);
}
#-----------------------------------------
#
#-------------------------------------------
sub print_edges {
    my $self=shift;
    
    my $edges=$self->{'graph'};

    foreach my $node1 (keys %{$edges}) {
	foreach my $node2 (keys %{$edges->{$node1}}) {
	    my $edge=$edges->{$node1}{$node2};
	    next unless defined $edge->{'ave_gap_size'};
	    print $node1."\t".$node2."\t".$edge->{'weigth'}."\t".$edge->{'stddev'}."\t".$edge->{'ave_gap_size'}."\t".$edge->{'count'}."\t".$edge->{'total_count'}."\n";
	}
    }

}
#-----------------
#
#--------------------
sub print_rep_node {

    my $self=shift;
    
    my $rep_node=$self->{'rep_node'};
    
    foreach my $node (keys %{$rep_node}) {
	print "Rep_node: $node\n";
    }

}
#------------------------------------------------------------------------------------------
# Purpose: find the most confidence path that order and oritentate the contigs
# Pre: the graph has been generated for the project
# Post: the most confidence path get printed out
# Note/Consideration: Greedy algorithm will be used.  The best edge will be considered first and
#   then only no conflict edges will be added.  In the case of sliding, backtracking will be needed
#   to identify a feasible path
#
# Algorithm:
#     initiate a path table with three column, 1. node, 2. neighbour node to NULL, 3. complement node to the complement
#     sort the edges in a descending order
#     foreach edges {
#         case 1: if both two nodes involved have no neighbor, update, next;
#         case 2: if only one has neighbor, one way sliding
#         case 3: if both have neighbor, two way sliding
#    }
#    find all nodes without neighbor
#    foreach such nodes {
#         if(! visited) {
#            depth travesal
#         }
#    }
#
#
#
#    one way sliding algorithm:
#         extract two lists that need to be merge(depth first travesal)
#         initilize current pointer of these two list at the node defined by the edge
#         ---recursive begin
#         merge_list() {
#             sum=0; #the sum of current gap2 plus contig
#             for(sum<gap1; sum+=(gap2+contig)) { #gap1 is the gap size of the current edge
#                if(can fit in) {
#                   insert the contig into the list
#                   update the current pointer of the two list to the next node(current->neighbor->complement)
#                   if(current pointer is NULL) #the end of the list
#                      update the table and return; #once one solution is found, take it
#                   else
#                      merge_list();     
#                   #---backtracking
#                   delete the contig from the list
#                   backdate the pointer
#                }
#             }
#          }
#
#
#    two way sliding algorithm:
#         extract two lists
#         split one of the list into two
#         merge_list with one half
#         merge_list with the other half
#         update the table if both succeed
#
#
# Data structure:
#-------------------------------------------------------------------------------------------
sub find_path {
    my($project)=shift;

    my ($edge, $contig, $node_name, $contig_name, $node1, $edge1, $node2, $edge2);
   # my %path_table=();
   # my @edges=();
    my %contigs=();

    $project->path_init(\%path_table, \@edges);

    #establish a hash table record all contig size involved in the path
    foreach $contig  (@{$project->{'contigs'}}) {
	$contig_name=$contig->name();
	$node_name=$contig_name."_left";
	if(exists $path_table{$node_name}) {
	    $contigs{$contig_name}=$contig->length();
	}
    }
    print "Finish loading contig hash\n" if ($VERBOSE);
    #sort the edges based on it weigth
    use vars qw($a $b); 
    @edges=sort( {$b->{'weigth'}<=>$a->{'weigth'}} @edges);

    #for debugging purpose
    if($ABORT) {
	print_path_table(\%path_table);
	print_edges_array(\@edges);
	return;
    }

    #add each edge into the path
    foreach $edge (@edges) {

	if($edge->{'weigth'} == 0) {  # reach the section where refered edges begins, a->b->c so a->c but its weight is count as 0
	    last;
	}

	$node1=$edge->{'node1'};
	$node2=$edge->{'node2'};
	#if two nodes are not already in the same path
	if( same_path($node1, $node2, \%path_table) ) {
		#do nothing. If add this link, cycle forms
	    next;
	}

	#the first case, non of the nodes involved are already connected
	if( (!defined $path_table{$node1}->{'next'}) && (! defined $path_table{$node2}->{'next'}) ) {
	    $path_table{$node1}->{'next'}=$node2;
	    $path_table{$node2}->{'next'}=$node1;
	    next;
	}
	
	#the second/third case, at least one node has next.
	use vars qw($path1 $path2 $flag);
	local ($path1, $path2);
	local $flag;
	$flag=0;
	if(defined $path_table{$node1}->{'next'}) {
	    $flag=1;
	}
	if(defined $path_table{$node2}->{'next'}) {
	    if( $flag == 1) {
		$flag=0; #both have next
	    } else {
		$flag=2; #only the second node has next
	    }
	}
	
	$path1=extract_path($node1, \%path_table);
	$path2=extract_path($node2, \%path_table);
	#the second case, one node is at the end of a path
	if($flag != 0) {
	    if($flag == 2) {  #swap path1 and 2
		my $temp=$path1;
		$path1=$path2;
		$path2=$temp;
	    }
	    if (&merge_uni_path(\%contigs, \@edges)) {
		&update_path_table($path1, \%path_table);
		next;
	    }
	    next;
	}

	#the third case, both nodes are in the middle of a path
	use vars qw($init_path1 $init_path2 $init_path2_next $end_count $fail);
	local $init_path1=$path1;
	local $init_path2=$path2;
	local $init_path2_next=$path2->{'next'};
	local $end_count=0;
	local $fail=0; #a flag for exit of recursion
	#if(&merge_bi_path(\%contigs, \@edges)) {
	  #  &update_path_table($path1, \%path_table);
	  #  next;
	#}
   }

    &print_path(\%path_table);

} #end of sub


#------------------------------------------------------------------
#
#----------------------------------------------------------------
sub extract_path {
    my($node_name, $path_table_ref)=@_;

    my($new_node_name, $new_node, $temp, $path_node, $path, $current_pos);

    #create a double link list/path with one node: the $node
    $path={'node'=>$node_name,
	   'comp'=>undef,
	   'next'=>undef
	   };

    $new_node_name=$node_name;
    $current_pos=$path;
    #depth first traversal
    while($new_node_name) {	
	#look for the comp ones first
	$new_node_name=$path_table_ref->{$new_node_name}->{'comp'};
	$new_node={'node'=>$new_node_name,
		   'comp'=>undef,
		   'next'=>undef
		   };
	$current_pos=&insert_comp_node($current_pos, $new_node);

	#look for neighbor first
	if( (defined $path_table_ref->{$new_node_name}->{'next'}) ){
	    $new_node_name=$path_table_ref->{$new_node_name}->{'next'};
	    $new_node={'node'=>$new_node_name,
		       'comp'=>undef,
		       'next'=>undef
		       };
	    $current_pos=&insert_next_node($current_pos, $new_node);
	} else {
	    $new_node_name=undef;
	}
    }

    #if the node is in the middle of a path
    if( (defined $path_table_ref->{$node_name}->{'next'}) ) {
	$new_node_name=$path_table_ref->{$node_name}->{'next'};
	$new_node={'node'=>$new_node_name,
		   'comp'=>undef,
		   'next'=>undef
		   };
        $current_pos=$path;
	$current_pos=&insert_next_node($current_pos, $new_node);

        #depth first traversal
        while($new_node_name) {	
	    #look for the comp ones first
	    $new_node_name=$path_table_ref->{$new_node_name}->{'comp'};
	    $new_node={'node'=>$new_node_name,
		       'comp'=>undef,
		       'next'=>undef
		       };
	    $current_pos=&insert_comp_node($current_pos, $new_node);

	    #look for neighbor first
	    if( (defined $path_table_ref->{$new_node_name}->{'next'}) ){
		$new_node_name=$path_table_ref->{$new_node_name}->{'next'};
		$new_node={'node'=>$new_node_name,
			   'comp'=>undef,
			   'next'=>undef
			   };
		$current_pos=&insert_next_node($current_pos, $new_node);
	    } else {
		$new_node_name=undef;
	    }
	}
    }
    return $path;
}


#----------------------------------------------------------
#
#---------------------------------------------------------
sub insert_comp_node {

    my($current_pos, $new_node)=@_;

    $current_pos->{'comp'}=$new_node;
    $new_node->{'comp'}=$current_pos;
    
    return $new_node;
}

#---------------------------------------------------------
#
#--------------------------------------------------------
sub insert_next_node {
    
    my($current_pos, $new_node)=@_;

    $current_pos->{'next'}=$new_node;
    $new_node->{'next'}=$current_pos;
    
    return $new_node;
}   

#------------------------------------------------------------
#
#---------------------------------------------------------
sub merge_uni_path {

    my($contig_len_ref, $edge_ref)= @_;

    my($sum, $node1, $node1_next, $node2, $contig1_len, $contig2_len, $gap1, $stddev1, $gap2, $stddev2);
 
    $node1=$path1->{'node'};
    $node2=$path2->{'node'};
    $node1_next=$path1->{'next'}->{'node'};

    ($gap2, $stddev2)=&get_gap_size($edge_ref, $node1, $node2);
    ($gap1, $stddev1)=&get_gap_size($edge_ref, $node1, $node1_next);

    $node1_next=~/(.+)_\S+$/;
    $contig1_len=$contig_len_ref->{$1};
    $node2=~/(.+)_\S+$/;
    $contig2_len=$contig_len_ref->{$1};

    my $orig_path1=$path1;
    my $current_path1_node=$path1;
    my $orig_path2=$path2;

    $sum=0;
    while($sum<($gap2+$stddev2)) {
	#if current contig of path2 can fit in the current gap of path1
	if( ($gap1+$sum+$stddev1)>=($gap2+$contig2_len-$stddev2) ) {

	    #push two new edge into the @edge.  This is essential for calculating new gap size later on
	    my $new_edge1={'node1'=>$current_path1_node->{'node'},
			   'node2'=>$path2->{'node'},
			   'weigth'=>0,
			   'stddev'=>$stddev2,
			   'size'=>($gap2-$sum)
		       };
	    my $new_edge2={'node1'=>$path2->{'comp'}->{'node'},
			   'node2'=>$current_path1_node->{'next'}->{'node'},
			   'weigth'=>0,
			   'stddev'=>$stddev1,
			   'size'=>($sum+$gap1-$gap2-$contig2_len)
		       };
	    push(@{$edge_ref}, $new_edge1);
	    push(@{$edge_ref}, $new_edge2);

	    #merge the contig of path2 into path1
	    my $new_path2=$path2->{'comp'}->{'next'};
	    my $new_path1=$path2->{'comp'};
	    $path2->{'comp'}->{'next'}=$current_path1_node->{'next'};  #order is important
	    $current_path1_node->{'next'}->{'next'}=$path2->{'comp'};
	    $current_path1_node->{'next'}=$path2;
	    $path2->{'next'}=$current_path1_node;

	    $path1=$new_path1;
	    $path2=$new_path2;
	    
	    if(! defined $path2) {  #the end of the path2 is reached successfully
		return 1;
	    } else {
		return 1 if &merge_uni_path($contig_len_ref, $edge_ref);
		return 0; #if fail, do not need to backtrack, return
	    }

	    #back tracking

	    pop (@{$edge_ref});
	    pop (@{$edge_ref});

	    $path1->{'comp'}->{'next'}->{'next'}=$path1->{'next'};
	    $path1->{'next'}->{'next'}=$path1->{'comp'}->{'next'};
	    $path2->{'next'}=$path1;
	    $path1->{'next'}=$path2;
	    $path1=$orig_path1;
	    $path2=$orig_path2;
	}
	$current_path1_node=$current_path1_node->{'next'}->{'comp'};
	$sum=$sum+$gap1+$contig1_len;

	#check to see if reach the end of the first path
	if(! defined $current_path1_node->{'next'}) {
	    	if( ($gap2+$stddev2)>=$sum ) {
		    $current_path1_node->{'next'}=$path2;
	       	    $path2->{'next'}=$current_path1_node;
		    return 1;
		} else {
		    return 0;
		}
	}

	$node1=$current_path1_node->{'node'};
	$node1_next=$current_path1_node->{'next'}->{'node'};

	($gap1, $stddev1)=&get_gap_size($edge_ref, $node1, $node1_next);

	$node1_next=~/(.+)_\S+$/;
	$contig1_len=$contig_len_ref->{$1};
    }

    #if reach this point, fail to merge two paths
    return 0;
}

#-----------------------------------------------------------
#
#----------------------------------------------------------
sub get_gap_size {
    my($edge_ref, $node1, $node2)=@_;
    
    my($count, $edge_node1, $edge_node2);
    for($count=0; $count<=$#{$edge_ref}; $count++) {
	$edge_node1=$edge_ref->[$count]->{'node1'};
	$edge_node2=$edge_ref->[$count]->{'node2'};
	if( (($node1 eq $edge_node1) && ($node2 eq $edge_node2)) || 
	    (($node2 eq $edge_node1) && ($node1 eq $edge_node2)) ) {
	    return ($edge_ref->[$count]->{'size'}, $edge_ref->[$count]->{'stddev'});
	}
    }
    return (0, 0);
}
    
#----------------------------------------------------
#
#-----------------------------------------------------
sub update_path_table {

    my ($path, $path_table_ref)=@_;
    my ($current_node, $current_node_name, $next_node_name);

    $current_node=$path;
    while($current_node) {	
	#look for the next ones first
	$current_node_name=$current_node->{'node'};
	$next_node_name=$current_node->{'next'}->{'node'};

	#look for neighbor first
	if(defined $next_node_name){
	    $path_table_ref->{$current_node_name}->{'next'}=$next_node_name;
	    $path_table_ref->{$next_node_name}->{'next'}=$current_node_name;
	} else {
	    last;
	}

	$current_node=$current_node->{'next'}->{'comp'};
    }  
    
    #go to the other direction
    $current_node=$path->{'comp'};
    while($current_node) {	
	#look for the next ones first
	$current_node_name=$current_node->{'node'};
	$next_node_name=$current_node->{'next'}->{'node'};

	#look for neighbor first
	if(defined $next_node_name){
	    $path_table_ref->{$current_node_name}->{'next'}=$next_node_name;
	    $path_table_ref->{$next_node_name}->{'next'}=$current_node_name;
	} else {
	    last;
	}

	$current_node=$current_node->{'next'}->{'comp'};
    }  
}




#----------------------------------------------------
#
#------------------------------------------------------
sub merge_bi_path {

    my($contig_len_ref, $edge_ref)= @_;

    my($sum, $node1, $node1_next, $node2, $contig1_len, $contig2_len, $gap1, $stddev1, $gap2, $stddev2);
 
    $node1=$path1->{'node'};
    $node2=$path2->{'node'};
    $node1_next=$path1->{'next'}->{'node'};

    ($gap2, $stddev1)=&get_gap_size($edge_ref, $node1, $node2);
    ($gap1, $stddev2)=&get_gap_size($edge_ref, $node1, $node1_next);

    $node1_next=~/(.+)_\S+$/;
    $contig1_len=$contig_len_ref->{$1};
    $node2=~/(.+)_\S+$/;
    $contig2_len=$contig_len_ref->{$1};

    my $orig_path1=$path1;
    my $current_path1_node=$path1;
    my $orig_path2=$path2;
    my $orig_path2_next=$path2->{'next'};

    $sum=0;
    while($sum<$gap2+$stddev2) {
	#if current contig of path2 can fit in the current gap of path1
	if( ($gap1+$sum+$stddev1)>=($gap2+$contig2_len-$stddev2) ) {

	    #push two new edge into the @edge.  This is essential for calculating new gap size later on
	    my $new_edge1={'node1'=>$current_path1_node->{'node'},
			   'node2'=>$path2->{'node'},
			   'weigth'=>0,
			   'stddev'=>$stddev2,
			   'size'=>($gap2-$sum)
		       };
	    my $new_edge2={'node1'=>$path2->{'comp'}->{'node'},
			   'node2'=>$current_path1_node->{'next'}->{'node'},
			   'weigth'=>0,
			   'stddev'=>$stddev1,
			   'size'=>($sum+$gap1-$gap2-$contig2_len)
		       };
	    push(@{$edge_ref}, $new_edge1);
	    push(@{$edge_ref}, $new_edge2);

	    #merge the contig of path2 into path1
	    my $new_path2=$path2->{'comp'}->{'next'};
	    my $new_path1=$path2->{'comp'};
	    $path2->{'comp'}->{'next'}=$current_path1_node->{'next'};
	    $current_path1_node->{'next'}->{'next'}=$path2->{'comp'};
	    $current_path1_node->{'next'}=$path2;
	    $path2->{'next'}=$current_path1_node;

	    $path1=$new_path1;
	    $path2=$new_path2;
	    
	    if(! defined $path2) {  #the end of the path2 is reached successfully
		if($end_count==0) { #the first time reach the end
		    $end_count++;
		    $path1=$init_path2;
		    $path2=$init_path2_next;
		    if (&merge_uni_path($contig_len_ref, $edge_ref)) {
			return 1;
		    } else {
			return 0;
		    }
		} else { #both ends have been reached 
		    return 1;
		}
	    } else {
		return 1 if &merge_bi_path($contig_len_ref, $edge_ref);
	    }

	    #back tracking

	    pop (@{$edge_ref});
	    pop (@{$edge_ref});

	    $path1->{'comp'}->{'next'}->{'next'}=$path1->{'next'};
	    $path1->{'next'}->{'next'}=$path1->{'comp'}->{'next'};
	    $path2->{'next'}=$path1;
	    $path1->{'next'}=$path2;
	    $path1->{'comp'}->{'next'}=$orig_path2_next;
	    $path1=$orig_path1;
	    $path2=$orig_path2;
	}
	$current_path1_node=$current_path1_node->{'next'}->{'comp'};
	$sum=$sum+$gap1+$contig1_len;

	#check to see if reach the end of the first path
	if(! defined $current_path1_node->{'next'}) {
	    	if( ($gap2+$stddev2>=$sum) ) {
		    $current_path1_node->{'next'}=$path2;
	       	    $path2->{'next'}=$current_path1_node;
		    if($end_count==0) {
			$end_count++;
			$path1=$init_path2;
			$path2=$init_path2_next;
			return 1 if &merge_uni_path($contig_len_ref, $edge_ref);
			return 0; #if fail to get the other side
		    } else {
			return 1;
		    }
		} else {
		    return 0;
		}
	}

	$node1=$current_path1_node->{'node'};
	$node1_next=$current_path1_node->{'next'}->{'node'};

	($gap1, $stddev1)=&get_gap_size($edge_ref, $node1, $node1_next);

	$node1_next=~/(.+)_\S+$/;
	$contig1_len=$contig_len_ref->{$1};
    }

    #if reach this point, fail to merge two paths
    return 0;

}



#--------------------------------------------------
#
#------------------------------------------------
sub print_path {
    my $path_table_ref=\%path_table;
    
    my($node, @heads);

    
    #reset all nodes in the $\%path_table to 0 if existed
    #find out all nodes without incoming nodes, 'next' undef
    @heads=();
    foreach $node (keys %{$path_table_ref}) {
	if( defined $path_table_ref->{$node}->{'visited'} ) {
	    $path_table_ref->{$node}->{'visited'}=0;
	}
	if( ! defined $path_table_ref->{$node}->{'next'} ) {
	    push(@heads, $node);
	}
    }

    foreach $node (@heads) {
	while(! defined $path_table_ref->{$node}->{'visited'} || $path_table_ref->{$node}->{'visited'}==0 ) {
	    $path_table_ref->{$node}->{'visited'}=1;
	    print "$node->";
	    $node=$path_table_ref->{$node}->{'comp'};
	    $path_table_ref->{$node}->{'visited'}=1;
	    print "$node->";
	    $node=$path_table_ref->{$node}->{'next'};
	    if(! defined $node) {
		print "end\n";
		last;
	    }
	}
    }
    
}

#---------------------------------
#get the end node name of a give path defined by
#the input node.  Look for its upstream side defined
#by looking at next node first then comp
#if the node is an isolated one, return itself
#----------------------------------
sub get_end_node {

    my $project=shift;
    my $node=shift;
    
    my $path_table_ref=\%path_table; 
    
    return $node unless exists $path_table_ref->{$node};
    my $new_node_name=$node;
    if( (defined $path_table_ref->{$node}->{'next'}) ) {
        #depth first traversal
        while(defined $path_table_ref->{$new_node_name}->{'next'}) {	
	    $new_node_name=$path_table_ref->{$new_node_name}->{'next'};
	    $new_node_name=$path_table_ref->{$new_node_name}->{'comp'};
	}
	return $new_node_name;
    } else {
	return $node;
    }
}


#--------------------------------------------------
#print all the path with the first path and the last
#path defined
#&print_path_with_order(firstnodename, lastnodename)
#------------------------------------------------
sub print_path_with_order {
    my $project=shift;
    my($firstnode, $lastnode)=@_;

    my $path_table_ref=\%path_table;    
    my($node, @heads);
	 
    #reset all nodes in the $\%path_table to 0 if existed
    #find out all nodes without incoming nodes, 'next' undef
    @heads=();
    foreach $node (keys %{$path_table_ref}) {
		  if( defined $path_table_ref->{$node}->{'visited'} ) {
				$path_table_ref->{$node}->{'visited'}=0;
		  }
		  if( ! defined $path_table_ref->{$node}->{'next'} ) {
				push(@heads, $node);
		  }
    }
    #print out the first path 
    if(exists $path_table_ref->{$firstnode}) {
		  $node=$firstnode;
		  while(1) {
				$path_table_ref->{$node}->{'visited'}=1;
				print "$node->";
				$node=$path_table_ref->{$node}->{'comp'};
				$path_table_ref->{$node}->{'visited'}=1;
				print "$node->";
				$node=$path_table_ref->{$node}->{'next'};
				if(! defined $node) {
					 print "end\n";
					 last;
				}
	}
    }
    #mark out the last path
    if(exists $path_table_ref->{$lastnode}) {
		  $node=$lastnode;
		  while($node) {
				$path_table_ref->{$node}->{'visited'}=1;
				$node=$path_table_ref->{$node}->{'comp'};
				$path_table_ref->{$node}->{'visited'}=1;
				$node=$path_table_ref->{$node}->{'next'};
		  }
    }
	 
    #print out middle paths
    foreach $node (@heads) {
		  while(! defined $path_table_ref->{$node}->{'visited'} || $path_table_ref->{$node}->{'visited'}==0 ) {
				$path_table_ref->{$node}->{'visited'}=1;
				print "$node->";
				$node=$path_table_ref->{$node}->{'comp'};
				$path_table_ref->{$node}->{'visited'}=1;
				print "$node->";
				$node=$path_table_ref->{$node}->{'next'};
				if(! defined $node) {
					 print "end\n";
					 last;
				}
		  }
    }
    #print out the last path
    if(exists $path_table_ref->{$lastnode} && ! &same_path($firstnode, $lastnode)) {
		  $node=$lastnode;
		  while($node) {
				print "$node->";
				$node=$path_table_ref->{$node}->{'comp'};
				print "$node->";
				$node=$path_table_ref->{$node}->{'next'};
		  }
		  print "end\n";
    }   
    
}

#-------------------------------------------------------------------
#print_path_with_node
#pre: the path table has be established
#post: the path that contain the node have been print out
#Note: can be called either with $project->& with only node name 
# or with both node name and path table reference 
#--------------------------------------------------------
sub print_path_with_node {
    my $project=$_[0];
    my $node_name=$_[1];
    my $fh=$_[2];
    my $path_table_ref=$_[3];
    $path_table_ref=\%path_table if(! defined $path_table_ref);
    my $edge_ref=$project->get_edges_array_ref();

    my ($contig_name, $node_name_tmp);
    my %contigs=();
    #establish a hash table record all contig size involved in the path
    foreach my $contig (@{$project->{'contigs'}}) {
	$contig_name=$contig->name();
	$node_name_tmp=$contig_name."_left";
	if(exists $path_table_ref->{$node_name_tmp}) {
	    $contigs{$contig_name}=$contig->length();
	}
    }

    #go to the end of the path
    my $new_node_name=$node_name;
    #if the node is in the middle of a path
    if( (defined $path_table_ref->{$node_name}->{'next'}) ) {
	$new_node_name=$path_table_ref->{$node_name}->{'next'};
        #depth first traversal
        while(defined $path_table_ref->{$new_node_name}->{'next'}) {	
	    #look for the comp ones first
	    $new_node_name=$path_table_ref->{$new_node_name}->{'comp'};
	    #look for neighbor first
	    if( (defined $path_table_ref->{$new_node_name}->{'next'}) ){
		$new_node_name=$path_table_ref->{$new_node_name}->{'next'};
	    } else {
		last;
	    }
	}
    }

    my $total_size=0;
    my $previous_node_name;
    #print out the path
    while(defined $new_node_name ) {
	print $fh "$new_node_name->";
	$new_node_name=~/(.+)_\S+$/; #get the contig size
	my $contig_len=$contigs{$1};
	$total_size += $contig_len;
	printf $fh ("%d->", $contig_len);
	$new_node_name=$path_table_ref->{$new_node_name}->{'comp'};
	print $fh "$new_node_name->";
	$previous_node_name=$new_node_name;
	$new_node_name=$path_table_ref->{$new_node_name}->{'next'};
	if(defined $new_node_name) {
	    my ($gap, $stddev)=&get_gap_size($edge_ref, $previous_node_name, $new_node_name);
	    printf $fh ("%d->", $gap);
	    $total_size += $gap;
	}
    }
    printf $fh ("end\tTOTAL SIZE: %d\n", $total_size);
}

    

#--------------------------------------------------------------------
#Purpose: Find out if two nodes are in the same path
#Pre: the path_table has been established, the name of two nodes entered
# together with the reference of the path_table, nodes can be anywhere in the path
#Post: return a Boolean, 1 indicates in the same path, 0 for not in
#Error: if some of the two nodes is not in the path_table at all, 0 return 
# together with a warning
#---------------------------------------------------------------------
sub same_path {
    my ($node1, $node2, $path_table_ref)=@_;
    
    $path_table_ref = \%path_table if(! defined $path_table_ref);

    my($current_node, $next_node, $temp);
    
    unless( (exists $path_table_ref->{$node1}) && (exists $path_table_ref->{$node2}) ) {
	print STDERR "$node1 or $node2 is not in the path_table\n" if ($VERBOSE);
	return 0;
    }

    $current_node=$node1;
    #depth first traversal
    while($current_node) {
	$path_table_ref->{$current_node}->{'visited'}=1;
	if($current_node eq $node2) {
	    print STDERR "node1 and node2 is at the same path\n" if ($VERBOSE);
	    #reset the visit into undef
	    foreach $temp (keys %{$path_table_ref}) {
		$path_table_ref->{$temp}->{'visited'}=undef;
	    }
	    return 1;
	}
	#look for neighbor first
	if( (defined $path_table_ref->{$current_node}->{'next'}) ){
	    $temp=$path_table_ref->{$current_node}->{'next'};
	    unless(defined $path_table_ref->{$temp}->{'visited'}) {
		$current_node=$temp;
		next;
	    }
	}
       
	#look for the comp ones if the neigbor fails
	$temp=$path_table_ref->{$current_node}->{'comp'};
	unless(defined $path_table_ref->{$temp}->{'visited'}) {
	    $current_node=$temp;
	    next;
	}

	#if both neighbor and next fail, that is the end of the path
	print STDERR "one end of the path is reached\n" if ($VERBOSE);
	$current_node=undef;
    }

    #look at if the node1 is at the middle of a path
    if( (defined $path_table_ref->{$node1}->{'next'}) ) {
	print STDERR "the node1 is at the middle of a path\n" if ($VERBOSE);
	$current_node= $path_table_ref->{$node1}->{'comp'};
	while($current_node) {
	    $path_table_ref->{$current_node}->{'visited'}=1;
	    if($current_node eq $node2) {
		print STDERR "node1 and node2 is at the same path\n" if ($VERBOSE);
		#reset the visit into undef
		foreach $temp (keys %{$path_table_ref}) {
		    $path_table_ref->{$temp}->{'visited'}=undef;
		}
		return 1;
	    }
	    #look for neighbor first
	    if( (defined $path_table_ref->{$current_node}->{'next'}) ){
		$temp=$path_table_ref->{$current_node}->{'next'};
		unless(defined $path_table_ref->{$temp}->{'visited'}) {
		    $current_node=$temp;
		    next;
		}
	    }
       
	    #look for the comp ones if the neigbor fails
	    $temp=$path_table_ref->{$current_node}->{'comp'};
	    unless(defined $path_table_ref->{$temp}->{'visited'}) {
		$current_node=$temp;
		next;
	    }

	    #if both neighbor and next fail, that is the end of the path
	    print STDERR "the second end of the path is reached\n" if ($VERBOSE);
	    $current_node=undef;
	}
    }
    # a clean up
    foreach $temp (keys %{$path_table_ref}) {
	$path_table_ref->{$temp}->{'visited'}=undef;
    }
    return 0;
}
	    
	    
	 
	   
#-----------------------------------------------------------------
#Purpose: initilize the path table and edges array based on the graph 
#Pre: The graph has been established
#Post: unique edges have been pushed into the @edges,
#      the path_table hash has been initilized
#---------------------------------------------------------
sub path_init{
    my($project)=shift;
    my($path_table_ref)=shift;
    my($edges_ref)=shift;

    #reset @edges and %path_table
    undef @{$edges_ref};
    undef %{$path_table_ref};

   #Initilize the path table and store all edges(unique ones) into an array
    #note: only contigs appears in the graph will be built into the table
    my($start_node, $end_node, $nodes_name, $comp_node);
    my %seen=(); #keep track of two nodes combination, only uniqe edge will be retain

    foreach $start_node (keys %{$project->{'graph'}}) {
	#updating the path table
	if(! exists $path_table_ref->{$start_node} ) {
	    $start_node=~/(\S+)_(\S+)$/;
	    if($2 eq 'left') {
		$comp_node=$1."_right";
	    } else {
		$comp_node=$1."_left";
	    }

	    $path_table_ref->{$start_node}={'next'=>undef,
					    'comp'=>$comp_node,
					    'visit'=>undef
					    };
	    $path_table_ref->{$comp_node}={'next'=>undef,
					   'comp'=>$start_node,
					   'visit'=>undef
					   };
	}

	#push the edges into the @edges
	foreach $end_node (keys %{$project->{'graph'}{$start_node}}) {
	    $nodes_name="$start_node+$end_node";
	    next if (exists $seen{$nodes_name});
	    my $current_node=$project->{'graph'}{$start_node}{$end_node};
	    next if ($current_node->{'weigth'}<1);
	    my $edge={'node1'=>$start_node,
		   'node2'=>$end_node,
		   'weigth'=>$current_node->{'weigth'},
		      'stddev'=>$current_node->{'stddev'},
		   'size'=>$current_node->{'ave_gap_size'}
		   };
	    push(@{$edges_ref}, $edge);

	    #update the path table
	    if(! exists $path_table_ref->{$end_node}) {
		$end_node=~/(\S+)_(\S+)$/;
		if($2 eq 'left') {
		    $comp_node=$1."_right";
		} else {
		    $comp_node=$1."_left";
		}

		$path_table_ref->{$end_node}={'next'=>undef,
					      'comp'=>$comp_node,
					      'visit'=>undef
					      };
		$path_table_ref->{$comp_node}={'next'=>undef,
					       'comp'=>$end_node,
					       'visit'=>undef
					       };
	    }

	    #update the seen hash for unique edges seen so far
	    $nodes_name="$start_node+$end_node";
	    $seen{$nodes_name}=1;
	    $nodes_name="$end_node+$start_node";
	    $seen{$nodes_name}=1;
	}
    }
}


############################
#print out the content of path_table, for debug use
##############################
sub print_path_table {
    my($path_table_ref)=shift;

    my $node;
    foreach $node (keys %$path_table_ref ) {
	my $value=$path_table_ref->{$node};
	print "$node\t";
	if(defined $value->{'next'}) { print $value->{'next'}."\t";}
	else { print "--\t"};
	
	print $value->{'comp'}."\t";
	if(defined $value->{'visit'}) { print $value->{'visit'}."\n";}
	else { print "--\n";}
    }
}

################################################
#print out the content of the edges array
#################################################
sub print_edges_array {
    my($edges_ref)=shift;

    my $edge;
    foreach $edge (@$edges_ref) {
	print $edge->{'node1'}."\t".$edge->{'node2'}."\t".$edge->{'weigth'}."\t".$edge->{'stddev'}."\t".$edge->{'size'}."\n";
    }
}



###################################################
#
####################################################
sub debug_path {
    my($project)=shift;
    my($infile)=shift;
    
    $project->set_verbose(1);

    my ($edge, $contig, $node_name, $contig_name, $node1, $edge1, $node2, $edge2);
    %path_table=();
    @edges=();
    my %contigs=();

    #read in the pathtable and edges from external file
    open (INFILE, "$infile");
    while(<INFILE>) {
	if(/^(\S+)\t(\S+)\n/) {
	    $contigs{$1}=$2;
	}
	if(/^(\S+)\t(\S+)\t(\S+)\t(\S+)\n/) {
	    $path_table{$1}={'next'=>$2,
			       'comp'=>$3,
			       'visit'=>$4
			       };
	    if($2 eq '--') {$path_table{$1}->{'next'}=undef;}
	    if($4 eq '--') {$path_table{$1}->{'visit'}=undef;}
	} elsif (/^(\S+)\t(\S+)\t(\S+)\t(\S+)\t(\S+)\n/) {
	    my $newedge={'node1'=>$1,
			 'node2'=>$2,
			 'weigth'=>$3,
			 'stddev'=>$4,
			 'size'=>$5
			 };
	    push (@edges, $newedge);
	}
    }
    close INFILE;
    #add each edge into the path
    foreach $edge (@edges) {

	if($edge->{'weigth'} == 0) {  # reach the section where refered edges begins, a->b->c so a->c but its weight is count as 0
	    last;
	}

	$node1=$edge->{'node1'};
	$node2=$edge->{'node2'};
	#if two nodes are not already in the same path
	if( same_path($node1, $node2, \%path_table) ) {
		#do nothing. If add this link, cycle forms
	    next;
	}

	#the first case, non of the nodes involved are already connected
	if( (!defined $path_table{$node1}->{'next'}) && (! defined $path_table{$node2}->{'next'}) ) {
	    $path_table{$node1}->{'next'}=$node2;
	    $path_table{$node2}->{'next'}=$node1;
	    next;
	}
	
	#the second/third case, at least one node has next.
	use vars qw($path1 $path2 $flag);
	local ($path1, $path2, $flag);
	$flag=0;
	if(defined $path_table{$node1}->{'next'}) {
	    $flag=1;
	}
	if(defined $path_table{$node2}->{'next'}) {
	    if( $flag == 1) {
		$flag=0;
	    } else {
		$flag=2;
	    }
	}
	
	$path1=extract_path($node1, \%path_table);
	$path2=extract_path($node2, \%path_table);
	#the second case, one node is at the end of a path
	if($flag != 0) {
	    if($flag == 2) {
		my $temp=$path1;
		$path1=$path2;
		$path2=$temp;
	    }
	    if (&merge_uni_path(\%contigs, \@edges)) {
		&update_path_table($path1, \%path_table);
		next;
	    }
	    next;
	}

	#the third case, both nodes are in the middle of a path
	use vars qw($init_path1 $init_path2 $end_count);
	local $init_path1=$path1;
	local $init_path2=$path2;
	local $end_count=0;
	if(&merge_bi_path(\%contigs, \@edges)) {
	    &update_path_table($path1, \%path_table);
	    next;
	}
   }

    &print_path();

} #end of sub
#######################################


# Put the contigs into biggest to smallest order.
sub sort_contigs_by_length {
    my($self)=shift;

    use vars qw($a $b); 
    @{$self->{'contigs'}}=sort({$b->length()<=>$a->length()}@{$self->{'contigs'}});
}

# Sort all of the contigs traces.
sub sort_traces_by_contig_start {
    my($self)=shift;
    my($contig);

    foreach $contig ($self->list_contigs()) {
	$contig->sort_traces_by_contig_start();
    }
}


1;

__END__


